import ohio.ext.pandas
import pandas as pd
import verboselogs, logging

logger = verboselogs.VerboseLogger(__name__)

from scipy import stats
import yaml
import psycopg2
from io import StringIO

from triage.component.architect.database_reflection import table_exists
from triage.component.catwalk.storage import ProjectStorage
from triage.component.catwalk.utils import save_db_objects


class CrosstabsConfigLoader:
    def __init__(self, config_path=None, config=None):
        self.output = {}
        self.models_list_query = None
        self.as_of_dates_query = None
        self.models_dates_join_query = None
        self.features_query = None
        self.predictions_query = None

        if config:
            config_to_load = config
        elif config_path:
            with open(config_path, "r") as stream:
                config_to_load = yaml.full_load(stream)
        else:
            raise ValueError("Either config_path or config are needed")
        self.__dict__.update(config_to_load)


# Later to be run against feature matrices of high- and low-risk entities.
# Each function receives two arguments: A dataframe of high-risk entities'
# features (as columns), and a dataframe of low-risk entities' features.
# The functions return either a Series, where the row index are the names of
# the feature columns, or a dataframe, again with the rows indexed by feature
# names.

high_risk_n = lambda hr, lr: pd.Series(index=hr.columns, data=hr.shape[0])
low_risk_n = lambda hr, lr: pd.Series(index=lr.columns, data=lr.shape[0])
high_risk_mean = lambda hr, lr: hr.mean(axis=0)
low_risk_mean = lambda hr, lr: lr.mean(axis=0)
high_risk_std = lambda hr, lr: hr.std(axis=0)
low_risk_std = lambda hr, lr: lr.std(axis=0)
hr_lr_ratio = lambda hr, lr: hr.mean(axis=0) / lr.mean(axis=0)
high_risk_support = lambda hr, lr: (hr > 0).sum(axis=0)
low_risk_support = lambda hr, lr: (lr > 0).sum(axis=0)
high_risk_support_pct = lambda hr, lr: round((hr > 0).sum(axis=0).astype(float) / len(hr), 3)
low_risk_support_pct = lambda hr, lr: round((lr > 0).sum(axis=0).astype(float) / len(lr), 3)

def hr_lr_ttest(hr, lr):
    """Returns the t-test (T statistic and p value), comparing the features for
    high- and low-risk entities."""
    res = stats.ttest_ind(
        hr.to_numpy(), lr.to_numpy(), axis=0, nan_policy="omit", equal_var=False
    )

    r0 = pd.Series(res[0], index=hr.columns)
    r1 = pd.Series(res[1], index=hr.columns)

    return pd.DataFrame({"ttest_T": r0, "ttest_p": r1})


# the crosstab functions are passed as a list of tuples;
# the first tuple element gives the name (or names) of the
# metric that the function calculates
crosstab_functions = [
    ("count_predicted_positive", high_risk_n),
    ("count_predicted_negative", low_risk_n),
    ("mean_predicted_positive", high_risk_mean),
    ("mean_predicted_negative", low_risk_mean),
    ("std_predicted_positive", high_risk_std),
    ("std_predicted_negative", low_risk_std),
    ("ratio_predicted_positive_over_predicted_negative", hr_lr_ratio),
    (["ttest_T", "ttest_p"], hr_lr_ttest),
]


def list_to_str(list_sql):
    str = "(" + ",".join(list_sql) + ")"
    return str


def populate_crosstabs_table(
    model_id,
    as_of_date,
    df,
    thresholds=None,
    crosstab_functions=crosstab_functions,
    push_to_db=True,
    return_df=False,
    engine=None,
    schema="results",
    table="crosstabs",
):
    """This function populates the results.crosstabs table.

    Args:
        entity_id_list (list): A list of entity_ids. Crosstabs will
                               be calculated using only predictions and
                               features of these entities. This list is
                               being passed to get_predictions_query and
                               get_features_query.
        crosstab_functions (dict): A dictionary of function names and
                               functions. The functions receive the feature
                               dataframes for entities that were predicted
                               postive or negative, respectively. Each function
                               returns a Series or DataFrame with rows now
                               indexed by the feature names, and columns
                               corresponding to the statistics that the
                               function calculated.
        thresholds (dict): A dictionary that maps column names to lists of
                           threshold values. The (name, threshold) pairs will
                           be applied to the dataframes with predictions to
                           split low- and high-risk entities.
        push_to_db (bool): If True, the crosstab table is being populated.
        return_df (bool): If True, a dataframe corresponding to the crosstabe
                          table will be returned.
        engine (SQLAlchemy engine): DB connection
        schema (str): Name of the schema to store results; defaults to 'results'
        table (str): Name of table to store results; defaults to 'crosstabs'

        Returns: A DataFrame, corresponding to results.crosstabs
    """
    logger.info(
        f"RUNNING populate_crosstabs_table for model_id {model_id} and as_of_date {as_of_date}",
    )
    if len(df) == 0:
        raise ValueError("No data could be fetched.")

    dfs = []

    # this will be useful later - the non-feature columns we'll grab
    model_cols = [
        "model_id",
        "as_of_date",
        "entity_id",
        "score",
        "rank_abs",
        "rank_pct",
        "label_value",
    ]

    # drop the all-null columns
    null_cols = df.columns[df.isnull().sum() == df.shape[0]]
    null_cols = [c for c in null_cols if c not in model_cols]

    if len(null_cols) > 0:
        logger.warning("Dropping the all-null columns: %s" % str(null_cols))
        df = df.drop(null_cols, 1)

    # if the dataframe isn't dummified, do it
    if object in df.dtypes.values or pd.Categorical.dtype in df.dtypes.values:
        to_dummify = [
            c
            for c in df.select_dtypes(include=["category", object])
            if c not in model_cols
        ]
        logger.debug("Dummifying the data")
        df = pd.get_dummies(df, columns=to_dummify, dummy_na=True)

    feat_cols = [c for c in df.columns if c not in model_cols]
    logger.debug("Iterating over thresholds to generate results...")
    for thres_type, thress in thresholds.items():
        for thres in thress:

            results = pd.DataFrame(index=feat_cols)
            logger.debug("split dataframe in high risk and lowriks")
            # split dataframe into high/low risk
            df_pred_pos = df.loc[df[thres_type] <= thres, feat_cols]
            df_pred_neg = df.loc[df[thres_type] > thres, feat_cols]
            logger.debug(f"len of hr: {len(df_pred_pos)} and lr: {len(df_pred_neg)}")
            for name, func in crosstab_functions:
                logger.debug(name)
                this_result = pd.DataFrame(func(df_pred_pos, df_pred_neg))
                if name in ["ttest_T", "ttest_p"]:
                    logger.debug("this_result:", this_result.shape)
                    logger.debug("Results:", results.shape)
                if not type(name) in [list, tuple]:
                    name = [name]
                # the metric name is coming from the crosstab_functions tuples
                this_result.columns = name
                results = results.join(this_result, how="outer")

            results = results.stack()
            results.index.names = ["feature_column", "metric"]
            results.name = "value"
            results = results.to_frame()

            results["model_id"] = model_id
            results["as_of_date"] = as_of_date
            results["threshold_unit"] = thres_type[-3:]
            results["threshold_value"] = thres

            dfs.append(results)

    df = pd.concat(dfs)
    if push_to_db:
        logger.debug("Pushing results to database...")
        df.reset_index().set_index(["model_id", "as_of_date", "metric"]).pg_copy_to(
            schema=schema, name=table, con=engine, if_exists="append"
        )
    if return_df:
        return df


def run_crosstabs(db_engine, crosstabs_config):
    crosstabs_query = """
        with models_list_query as (
        {models_list_query}
        ), as_of_dates_query as (
        {as_of_dates_query}
        ),models_dates_join_query as (
        {models_dates_join_query}
        ),features_query as (
        {features_query}
        ), predictions_query as (
        {predictions_query}
        )
        select * from predictions_query
        left join features_query f using (model_id,entity_id, as_of_date)
        """.format(
        models_list_query=crosstabs_config.models_list_query,
        as_of_dates_query=crosstabs_config.as_of_dates_query,
        models_dates_join_query=crosstabs_config.models_dates_join_query,
        features_query=crosstabs_config.features_query,
        predictions_query=crosstabs_config.predictions_query,
    )
    if len(crosstabs_config.entity_id_list) > 0:
        crosstabs_query += " where entity_id=ANY('{%s}') " % ", ".join(
            map(str, crosstabs_config.entity_id_list)
        )
    crosstabs_query += "  order by model_id, as_of_date, rank_abs asc;"
    df = pd.read_sql(crosstabs_query, db_engine)
    if len(df) == 0:
        raise ValueError("No data could be fetched.")
    groupby_obj = df.groupby(["model_id", "as_of_date"])
    for group, values in groupby_obj:
        df_modelid_asofdate = groupby_obj.get_group(group)
        res = populate_crosstabs_table(
            model_id=group[0],
            as_of_date=group[1],
            df=df_modelid_asofdate,
            thresholds=crosstabs_config.thresholds,
            crosstab_functions=crosstab_functions,
            push_to_db=True,
            return_df=False,
            engine=db_engine,
            schema=crosstabs_config.output["schema"],
            table=crosstabs_config.output["table"],
        )


def run_crosstabs_from_matrix(db_engine, project_path, model_id, threshold_type, threshold, matrix_uuid=None, push_to_db=True, table_schema='test_results', table_name='crosstabs', return_as_dataframe=True, replace=False):
    """ Calculate crosstabs for a model based on the matrix. 
        
        Args: 
            db_engine: Database engine
            project_path (str): Path where the experiment artifacts (models and matrices) are stored
            thresholds (Dict{str: Union[float, int}]): A dictionary that maps threhold type to the threshold
                                                    The threshold type can be one of the rank columns in the test_results.predictions_table
           
            matrix_uuid (str, optional): To run crosstabs for a different matrix than the validation matrix from the experiment

            push_to_db (bool, optional): Whether to write the results to the database. Defaults to True
            
            table_schame (str, optional): Database schema to store the crosstabs table. Defaults to `test_results`
            table_name (str, optional): Table name to use. Defaults to `crosstabs`. If the table exists, results are appended

        return:
            Dataframe of crosstabs
    """
    
    logging.info('Fetching predictions')
    
    where_clause = f'where model_id = {model_id}'
    
    if matrix_uuid:
        where_clause += f" and matrix_uuid = '{matrix_uuid}'"
    
    q = f'''
        select 
        matrix_uuid,
        entity_id,
        as_of_date,
        score, 
        label_value,
        test_label_timespan,
        {threshold_type},
        case when {threshold_type} <= {threshold} then 1 else 0 end as high_risk
        from test_results.predictions p
        {where_clause}
    '''
    
    id_columns = ['entity_id', 'as_of_date']
    
    predictions = pd.read_sql(q, db_engine).set_index(id_columns)
    
    if predictions.empty:
        logging.error('No predictions were found for the model and matrix')
        return 

    # The same model can have multiple test matrices. If not specified, we calculate crosstabs for all matrices
    matrices = [matrix_uuid]
    if matrix_uuid is None:
        matrices = predictions.matrix_uuid.unique()
        
    logging.info(f'Matrices for crosstabs: {matrices}')
    
    crosstab_tasks = list() # store a list of matrix uuids we need crosstabs for
        
    if table_exists(f'{table_schema}.{table_name}', db_engine):
        logging.info(f'The {table_schema}.{table_name} table exists. Checking for existing crosstabs')
        
        for matrix_uuid in matrices: 
            df = pd.read_sql(f'''
                    select 1 from {table_schema}.{table_name}
                    where model_id = {model_id}
                    and matrix_uuid = '{matrix_uuid}'
                    and threshold_type = '{threshold_type}'
                    and threshold = {threshold}
                ''', db_engine)
            
            if not df.empty:
                if replace: 
                    logging.info(f'Exsiting crosstabs found for model {model_id} and matrix {matrix_uuid}. Replace is True. Deleting.')
                    q = f'''
                        delete from {table_schema}.{table_name}
                        where model_id = {model_id}
                        and matrix_uuid = '{matrix_uuid}'
                        and threshold_type = '{threshold_type}'
                        and threshold = {threshold}
                    '''
                    
                    db_engine.execute(q)
                                        
                else:
                    logging.info(f'Existing crosstabs found for model {model_id} and matrix {matrix_uuid}. Replace flag is not set. Skipping')
                    continue
            
            logging.info(f'Crosstabs task added for matrix {matrix_uuid}')
            crosstab_tasks.append(matrix_uuid)
            
    else:
        logging.info(f'{table_schema}.{table_name} does not exist. Calculating crostabs for all matrices')
        crosstab_tasks = matrices
            
    if len(crosstab_tasks) == 0:
        logging.warning('Crosstabs calculation not needed. Exiting.')
        return
    
    project_storage = ProjectStorage(project_path)
    matrix_storage_engine = project_storage.matrix_storage_engine()
    
    results = list()
    for m in crosstab_tasks:
        matrix = matrix_storage_engine.get_store(matrix_uuid=m).design_matrix
        
        feature_names = matrix.columns
        
        matrix = matrix.join(predictions[predictions['matrix_uuid'] == m], how='left')
        
        # topk_msk = matrix[threshold_type] <= threshold
        topk_msk = matrix['high_risk'] == 1
        
        positives = matrix[topk_msk][feature_names]
        negatives = matrix[~topk_msk][feature_names]
        
        for name, function in crosstab_functions:
            logging.info(f'Calculating {name}')
            
            if name == ["ttest_T", "ttest_p"]:
                res = function(positives, negatives)

                this_result = res.drop(columns=['ttest_p'])
                
                # TODO: This is very hacky and ugly
                this_result.rename(columns={'ttest_T': 0}, inplace=True)
                this_result['metric'] = 'ttest_T'
                results.append(this_result)
                
                this_result = res.drop(columns=['ttest_T'])
                
                # TODO: This is very hacky and ugly
                this_result.rename(columns={'ttest_p': 0}, inplace=True)
                this_result['metric'] = 'ttest_p'
                
                results.append(this_result)
            
                continue
                
            this_result = pd.DataFrame(function(positives, negatives))
            
            this_result['metric'] = name
                
            results.append(this_result)
                        
    results = pd.concat(results).reset_index()
    results.rename(columns={'index': 'feature', 0:'value'}, inplace=True)
    results['threshold_type'] = threshold_type
    results['threshold'] = threshold
    results['model_id'] = model_id
    results['matrix_uuid'] = matrix_uuid
    
    if push_to_db:
        logging.info(f'Pushing the results to the database, {len(results)} rows')
                
        results.set_index(
            ['model_id', 'matrix_uuid', 'feature', 'metric', 'threshold_type', 'threshold'],
            inplace=True
        )
        
        results = results.reset_index()
        
        if not table_exists(f'{table_schema}.{table_name}', db_engine):
            q = f'''
                create schema if not exists {table_schema};
                
                create table {table_schema}.{table_name} (
                  model_id INTEGER,
                  matrix_uuid TEXT,
                  feature TEXT,
                  metric TEXT,
                  threshold_type TEXT,
                  threshold FLOAT,
                  value FLOAT  
                );
            
            '''
            # q = _generate_create_table_sql_statement_from_df(results, f'{table_schema}.{table_name}')
            db_engine.execute(q)
        
        conn = db_engine.raw_connection()
        cursor = conn.cursor()
        
        buffer = StringIO()
        results.to_csv(buffer, index=False, header=False)
        buffer.seek(0)
        
        columns = ', '.join(results.columns)
        print(columns)
        cursor.copy_expert(f"COPY {table_schema}.{table_name} ({columns}) FROM STDIN WITH CSV", buffer)
        # results.to_sql(con=db_engine, schema=table_schema, name=table_name, if_exists='append')
        conn.commit()
        cursor.close()
        conn.close()
        
    return results
        
                
        


"""add label_value to prodcution table

Revision ID: 264786a9fe85
Revises: 1b990cbc04e4
Create Date: 2019-02-26 13:17:05.365654

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '264786a9fe85'
down_revision = '1b990cbc04e4'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_table("list_predictions", schema="production")
    op.create_table(
        "list_predictions",
        sa.Column("model_id", sa.Integer(), nullable=False),
        sa.Column("entity_id", sa.BigInteger(), nullable=False),
        sa.Column("as_of_date", sa.DateTime(), nullable=False),
        sa.Column("score", sa.Numeric(), nullable=True),
        sa.Column('label_value', sa.Integer, nullable=True),
        sa.Column("rank_abs", sa.Integer(), nullable=True),
        sa.Column("rank_pct", sa.Float(), nullable=True),
        sa.Column("matrix_uuid", sa.Text(), nullable=True),
        sa.Column("test_label_timespan", sa.Interval(), nullable=True),
        sa.ForeignKeyConstraint(["model_id"], ["triage_metadata.models.model_id"]),
        sa.PrimaryKeyConstraint("model_id", "entity_id", "as_of_date"),
        schema="production",
    )


def downgrade():
    op.drop_table("list_predictions", schema="production")
    op.create_table(
        "list_predictions",
        sa.Column("model_id", sa.Integer(), nullable=False),
        sa.Column("entity_id", sa.BigInteger(), nullable=False),
        sa.Column("as_of_date", sa.DateTime(), nullable=False),
        sa.Column("score", sa.Numeric(), nullable=True),
        sa.Column("rank_abs", sa.Integer(), nullable=True),
        sa.Column("rank_pct", sa.Float(), nullable=True),
        sa.Column("matrix_uuid", sa.Text(), nullable=True),
        sa.Column("test_label_timespan", sa.Interval(), nullable=True),
        sa.ForeignKeyConstraint(["model_id"], ["triage_metadata.models.model_id"]),
        sa.PrimaryKeyConstraint("model_id", "entity_id", "as_of_date"),
        schema="results",
    )




import datetime
from unittest import TestCase, mock
import pandas as pd
import testing.postgresql

from contextlib import contextmanager

from triage import create_engine
from triage.component.catwalk.utils import filename_friendly_hash
from triage.component.architect.feature_group_creator import FeatureGroup
from triage.component.architect.builders import MatrixBuilder
from triage.component.catwalk.db import ensure_db
from triage.component.catwalk.storage import ProjectStorage
from triage.component.results_schema.schema import Matrix

from .utils import (
    create_schemas,
    create_entity_date_df,
    convert_string_column_to_date,
    TemporaryDirectory,
)

from tests.utils import matrix_metadata_creator


# make some fake features data

states = [
    [0, "2016-01-01", False],
    [0, "2016-02-01", False],
    [0, "2016-03-01", False],
    [0, "2016-04-01", False],
    [0, "2016-05-01", False],
    [0, "2016-06-01", True],
    [1, "2016-01-01", False],
    [1, "2016-02-01", False],
    [1, "2016-03-01", False],
    [1, "2016-04-01", False],
    [1, "2016-05-01", False],
    [2, "2016-01-01", False],
    [2, "2016-02-01", True],
    [2, "2016-03-01", False],
    [2, "2016-04-01", True],
    [2, "2016-05-01", False],
    [3, "2016-01-01", False],
    [3, "2016-02-01", True],
    [3, "2016-03-01", False],
    [3, "2016-04-01", True],
    [3, "2016-05-01", False],
    [4, "2016-01-01", True],
    [4, "2016-02-01", True],
    [4, "2016-03-01", True],
    [4, "2016-04-01", True],
    [4, "2016-05-01", True],
    [5, "2016-01-01", False],
    [5, "2016-02-01", False],
    [5, "2016-03-01", False],
    [5, "2016-04-01", False],
    [5, "2016-05-01", False],
]

features0_pre = [
    [0, "2016-01-01", 2, 1],
    [1, "2016-01-01", 1, 2],
    [0, "2016-02-01", 2, 3],
    [1, "2016-02-01", 2, 4],
    [0, "2016-03-01", 3, 3],
    [1, "2016-03-01", 3, 4],
    [0, "2016-04-01", 4, 3],
    [1, "2016-05-01", 5, 4],
]

features1_pre = [
    [2, "2016-01-01", 1, 1],
    [3, "2016-01-01", 1, 2],
    [2, "2016-02-01", 2, 3],
    [3, "2016-02-01", 2, 2],
    [0, "2016-03-01", 3, 3],
    [1, "2016-03-01", 3, 4],
    [2, "2016-03-01", 3, 3],
    [3, "2016-03-01", 3, 4],
    [3, "2016-03-01", 3, 4],
    [0, "2016-03-01", 3, 3],
    [4, "2016-03-01", 1, 4],
    [5, "2016-03-01", 2, 4],
]

# collate will ensure every entity/date combination in the state
# table have an imputed value in the features table, so ensure
# this is true for our test (filling with 9's):
f0_dict = {(r[0], r[1]): r for r in features0_pre}
f1_dict = {(r[0], r[1]): r for r in features1_pre}

for rec in states:
    ent_dt = (rec[0], rec[1])
    f0_dict[ent_dt] = f0_dict.get(ent_dt, list(ent_dt + (9, 9)))
    f1_dict[ent_dt] = f1_dict.get(ent_dt, list(ent_dt + (9, 9)))

features0 = sorted(f0_dict.values(), key=lambda x: (x[1], x[0]))
features1 = sorted(f1_dict.values(), key=lambda x: (x[1], x[0]))

features_tables = [features0, features1]

# make some fake labels data

labels = [
    [0, "2016-02-01", "1 month", "booking", "binary", 0],
    [0, "2016-03-01", "1 month", "booking", "binary", 0],
    [0, "2016-04-01", "1 month", "booking", "binary", 0],
    [0, "2016-05-01", "1 month", "booking", "binary", 1],
    [0, "2016-01-01", "1 month", "ems", "binary", 0],
    [0, "2016-02-01", "1 month", "ems", "binary", 0],
    [0, "2016-03-01", "1 month", "ems", "binary", 0],
    [0, "2016-04-01", "1 month", "ems", "binary", 0],
    [0, "2016-05-01", "1 month", "ems", "binary", 0],
    [1, "2016-01-01", "1 month", "booking", "binary", 0],
    [1, "2016-02-01", "1 month", "booking", "binary", 0],
    [1, "2016-03-01", "1 month", "booking", "binary", 0],
    [1, "2016-04-01", "1 month", "booking", "binary", 0],
    [1, "2016-05-01", "1 month", "booking", "binary", 1],
    [1, "2016-01-01", "1 month", "ems", "binary", 0],
    [1, "2016-02-01", "1 month", "ems", "binary", 0],
    [1, "2016-03-01", "1 month", "ems", "binary", 0],
    [1, "2016-04-01", "1 month", "ems", "binary", 0],
    [1, "2016-05-01", "1 month", "ems", "binary", 0],
    [2, "2016-01-01", "1 month", "booking", "binary", 0],
    [2, "2016-02-01", "1 month", "booking", "binary", 0],
    [2, "2016-03-01", "1 month", "booking", "binary", 1],
    [2, "2016-04-01", "1 month", "booking", "binary", 0],
    [2, "2016-05-01", "1 month", "booking", "binary", 1],
    [2, "2016-01-01", "1 month", "ems", "binary", 0],
    [2, "2016-02-01", "1 month", "ems", "binary", 0],
    [2, "2016-03-01", "1 month", "ems", "binary", 0],
    [2, "2016-04-01", "1 month", "ems", "binary", 0],
    [2, "2016-05-01", "1 month", "ems", "binary", 1],
    [3, "2016-01-01", "1 month", "booking", "binary", 0],
    [3, "2016-02-01", "1 month", "booking", "binary", 0],
    [3, "2016-03-01", "1 month", "booking", "binary", 1],
    [3, "2016-04-01", "1 month", "booking", "binary", 0],
    [3, "2016-05-01", "1 month", "booking", "binary", 1],
    [3, "2016-01-01", "1 month", "ems", "binary", 0],
    [3, "2016-02-01", "1 month", "ems", "binary", 0],
    [3, "2016-03-01", "1 month", "ems", "binary", 0],
    [3, "2016-04-01", "1 month", "ems", "binary", 1],
    [3, "2016-05-01", "1 month", "ems", "binary", 0],
    [4, "2016-01-01", "1 month", "booking", "binary", 1],
    [4, "2016-02-01", "1 month", "booking", "binary", 0],
    [4, "2016-03-01", "1 month", "booking", "binary", 0],
    [4, "2016-04-01", "1 month", "booking", "binary", 0],
    [4, "2016-05-01", "1 month", "booking", "binary", 0],
    [4, "2016-01-01", "1 month", "ems", "binary", 0],
    [4, "2016-02-01", "1 month", "ems", "binary", 1],
    [4, "2016-03-01", "1 month", "ems", "binary", 0],
    [4, "2016-04-01", "1 month", "ems", "binary", 1],
    [4, "2016-05-01", "1 month", "ems", "binary", 1],
    [5, "2016-01-01", "1 month", "booking", "binary", 1],
    [5, "2016-02-01", "1 month", "booking", "binary", 0],
    [5, "2016-03-01", "1 month", "booking", "binary", 0],
    [5, "2016-04-01", "1 month", "booking", "binary", 0],
    [5, "2016-05-01", "1 month", "booking", "binary", 0],
    [5, "2016-01-01", "1 month", "ems", "binary", 0],
    [5, "2016-02-01", "1 month", "ems", "binary", 1],
    [5, "2016-03-01", "1 month", "ems", "binary", 0],
    [5, "2016-04-01", "1 month", "ems", "binary", 0],
    [5, "2016-05-01", "1 month", "ems", "binary", 0],
    [0, "2016-02-01", "3 month", "booking", "binary", 0],
    [0, "2016-03-01", "3 month", "booking", "binary", 0],
    [0, "2016-04-01", "3 month", "booking", "binary", 0],
    [0, "2016-05-01", "3 month", "booking", "binary", 1],
    [0, "2016-01-01", "3 month", "ems", "binary", 0],
    [0, "2016-02-01", "3 month", "ems", "binary", 0],
    [0, "2016-03-01", "3 month", "ems", "binary", 0],
    [0, "2016-04-01", "3 month", "ems", "binary", 0],
    [0, "2016-05-01", "3 month", "ems", "binary", 0],
    [1, "2016-01-01", "3 month", "booking", "binary", 0],
    [1, "2016-02-01", "3 month", "booking", "binary", 0],
    [1, "2016-03-01", "3 month", "booking", "binary", 0],
    [1, "2016-04-01", "3 month", "booking", "binary", 0],
    [1, "2016-05-01", "3 month", "booking", "binary", 1],
    [1, "2016-01-01", "3 month", "ems", "binary", 0],
    [1, "2016-02-01", "3 month", "ems", "binary", 0],
    [1, "2016-03-01", "3 month", "ems", "binary", 0],
    [1, "2016-04-01", "3 month", "ems", "binary", 0],
    [1, "2016-05-01", "3 month", "ems", "binary", 0],
    [2, "2016-01-01", "3 month", "booking", "binary", 0],
    [2, "2016-02-01", "3 month", "booking", "binary", 0],
    [2, "2016-03-01", "3 month", "booking", "binary", 1],
    [2, "2016-04-01", "3 month", "booking", "binary", 0],
    [2, "2016-05-01", "3 month", "booking", "binary", 1],
    [2, "2016-01-01", "3 month", "ems", "binary", 0],
    [2, "2016-02-01", "3 month", "ems", "binary", 0],
    [2, "2016-03-01", "3 month", "ems", "binary", 0],
    [2, "2016-04-01", "3 month", "ems", "binary", 0],
    [2, "2016-05-01", "3 month", "ems", "binary", 1],
    [3, "2016-01-01", "3 month", "booking", "binary", 0],
    [3, "2016-02-01", "3 month", "booking", "binary", 0],
    [3, "2016-03-01", "3 month", "booking", "binary", 1],
    [3, "2016-04-01", "3 month", "booking", "binary", 0],
    [3, "2016-05-01", "3 month", "booking", "binary", 1],
    [3, "2016-01-01", "3 month", "ems", "binary", 0],
    [3, "2016-02-01", "3 month", "ems", "binary", 0],
    [3, "2016-03-01", "3 month", "ems", "binary", 0],
    [3, "2016-04-01", "3 month", "ems", "binary", 1],
    [3, "2016-05-01", "3 month", "ems", "binary", 0],
    [3, "2016-05-01", "3 month", "ems", "binary", 0],
    [4, "2016-01-01", "3 month", "booking", "binary", 0],
    [4, "2016-02-01", "3 month", "booking", "binary", 0],
    [4, "2016-03-01", "3 month", "booking", "binary", 1],
    [4, "2016-04-01", "3 month", "booking", "binary", 0],
    [4, "2016-05-01", "3 month", "booking", "binary", 1],
    [4, "2016-01-01", "3 month", "ems", "binary", 0],
    [4, "2016-02-01", "3 month", "ems", "binary", 0],
    [4, "2016-03-01", "3 month", "ems", "binary", 0],
    [4, "2016-04-01", "3 month", "ems", "binary", 0],
    [4, "2016-05-01", "3 month", "ems", "binary", 1],
    [5, "2016-01-01", "3 month", "booking", "binary", 0],
    [5, "2016-02-01", "3 month", "booking", "binary", 0],
    [5, "2016-03-01", "3 month", "booking", "binary", 1],
    [5, "2016-04-01", "3 month", "booking", "binary", 0],
    [5, "2016-05-01", "3 month", "booking", "binary", 1],
    [5, "2016-01-01", "3 month", "ems", "binary", 0],
    [5, "2016-02-01", "3 month", "ems", "binary", 0],
    [5, "2016-03-01", "3 month", "ems", "binary", 0],
    [5, "2016-04-01", "3 month", "ems", "binary", 1],
    [5, "2016-05-01", "3 month", "ems", "binary", 0],
]

label_name = "booking"
label_type = "binary"

db_config = {
    "features_schema_name": "features",
    "labels_schema_name": "labels",
    "labels_table_name": "labels",
    "cohort_table_name": "cohort",
    "triage_metadata": "triage_metadata",
}

experiment_hash = None


@contextmanager
def get_matrix_storage_engine():
    with TemporaryDirectory() as temp_dir:
        yield ProjectStorage(temp_dir).matrix_storage_engine()


def test_make_entity_date_table():
    """Test that the make_entity_date_table function contains the correct
    values.
    """
    dates = [
        datetime.datetime(2016, 1, 1, 0, 0),
        datetime.datetime(2016, 2, 1, 0, 0),
        datetime.datetime(2016, 3, 1, 0, 0),
    ]

    # make a dataframe of entity ids and dates to test against
    ids_dates = create_entity_date_df(
        labels=labels,
        states=states,
        as_of_dates=dates,
        label_name="booking",
        label_type="binary",
        label_timespan="1 month",
    )

    with testing.postgresql.Postgresql() as postgresql:
        # create an engine and generate a table with fake feature data
        engine = create_engine(postgresql.url())
        #ensure_db(engine)
        create_schemas(
            engine=engine, features_tables=features_tables, labels=labels, states=states
        )

        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                experiment_hash=experiment_hash,
                engine=engine,
            )
            engine.execute("CREATE TABLE features.tmp_entity_date (a int, b date);")
            # call the function to test the creation of the table
            entity_date_table_name = builder.make_entity_date_table(
                as_of_times=dates,
                label_type="binary",
                label_name="booking",
                state="active",
                matrix_uuid="my_uuid",
                matrix_type="train",
                label_timespan="1 month",
            )

            # read in the table
            result = pd.read_sql(
                "select * from features.{} order by entity_id, as_of_date".format(
                    entity_date_table_name
                ),
                engine,
            )
            # compare the table to the test dataframe
            test = result == ids_dates
            assert test.all().all()

def test_make_entity_date_table_include_missing_labels():
    """Test that the make_entity_date_table function contains the correct
    values.
    """
    dates = [
        datetime.datetime(2016, 1, 1, 0, 0),
        datetime.datetime(2016, 2, 1, 0, 0),
        datetime.datetime(2016, 3, 1, 0, 0),
        datetime.datetime(2016, 6, 1, 0, 0),
    ]

    # same as the other make_entity_date_label test except there is an extra date, 2016-06-01
    # entity 0 is included in this date via the states table, but has no label

    # make a dataframe of entity ids and dates to test against
    ids_dates = create_entity_date_df(
        labels=labels,
        states=states,
        as_of_dates=dates,
        label_name="booking",
        label_type="binary",
        label_timespan="1 month",
    )
    # this line adds the new entity-date combo as an expected one
    ids_dates = ids_dates.append(
        {"entity_id": 0, "as_of_date": datetime.date(2016, 6, 1)}, ignore_index=True
    )

    with testing.postgresql.Postgresql() as postgresql:
        # create an engine and generate a table with fake feature data
        engine = create_engine(postgresql.url())
        #ensure_db(engine)
        create_schemas(
            engine=engine, features_tables=features_tables, labels=labels, states=states
        )

        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                experiment_hash=experiment_hash,
                include_missing_labels_in_train_as=False,
                engine=engine,
            )
            engine.execute("CREATE TABLE features.tmp_entity_date (a int, b date);")
            # call the function to test the creation of the table
            entity_date_table_name = builder.make_entity_date_table(
                as_of_times=dates,
                label_type="binary",
                label_name="booking",
                state="active",
                matrix_uuid="my_uuid",
                matrix_type="train",
                label_timespan="1 month",
            )

            # read in the table
            result = pd.read_sql(
                "select * from features.{} order by entity_id, as_of_date".format(
                    entity_date_table_name
                ),
                engine,
            )

            # compare the table to the test dataframe
            assert sorted(result.values.tolist()) == sorted(ids_dates.values.tolist())


class TestMergeFeatureCSVs(TestCase):
    def test_feature_load_queries(self):
        """Tests if the number of queries for getting the features are the same as the number of feature tables in
        the feature schema.
        """
        
        dates = [
            datetime.datetime(2016, 1, 1, 0, 0),
            datetime.datetime(2016, 2, 1, 0, 0),
            datetime.datetime(2016, 3, 1, 0, 0),
            datetime.datetime(2016, 6, 1, 0, 0),
        ]

        features = [["f1", "f2"], ["f3", "f4"]]

        # create an engine and generate a table with fake feature data
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            #ensure_db(engine)
            create_schemas(engine, features_tables, labels, states)

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    experiment_hash=experiment_hash,
                    engine=engine,
                    include_missing_labels_in_train_as=False,
                )

                # make the entity-date table
                entity_date_table_name = builder.make_entity_date_table(
                    as_of_times=dates,
                    label_type="binary",
                    label_name="booking",
                    state="active",
                    matrix_type="train",
                    matrix_uuid="1234",
                    label_timespan="1m",
                )

                feature_dictionary = {
                    f"features{i}": feature_list
                    for i, feature_list in enumerate(features)
                }

                result = builder.feature_load_queries(
                    feature_dictionary=feature_dictionary, 
                    entity_date_table_name=entity_date_table_name
                )
                
                # lenght of the list should be the number of tables in feature schema
                assert len(result) == len(features)


    def test_stitch_csvs(self):
        """Tests if all the features and label were joined correctly in the csv
        """
        dates = [
            datetime.datetime(2016, 1, 1, 0, 0),
            datetime.datetime(2016, 2, 1, 0, 0),
            datetime.datetime(2016, 3, 1, 0, 0),
            datetime.datetime(2016, 6, 1, 0, 0),
        ]

        features = [["f1", "f2"], ["f3", "f4"]]

        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            #ensure_db(engine)
            create_schemas(
                engine=engine, features_tables=features_tables, labels=labels, states=states
            )

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    experiment_hash=experiment_hash,
                    engine=engine,
                )

                feature_dictionary = {
                    f"features{i}": feature_list
                    for i, feature_list in enumerate(features)
                }

                # make the entity-date table
                entity_date_table_name = builder.make_entity_date_table(
                    as_of_times=dates,
                    label_type="binary",
                    label_name="booking",
                    state="active",
                    matrix_type="train",
                    matrix_uuid="1234",
                    label_timespan="1 month",
                )

                feature_queries = builder.feature_load_queries(
                    feature_dictionary=feature_dictionary,
                    entity_date_table_name=entity_date_table_name
                )

                label_query = builder.label_load_query(
                    label_name="booking",
                    label_type="binary",
                    entity_date_table_name=entity_date_table_name,
                    label_timespan='1 month'
                )

                matrix_store = matrix_storage_engine.get_store("1234")
                
                result = builder.stitch_csvs(
                    features_queries=feature_queries,
                    label_query=label_query,
                    matrix_store=matrix_store,
                    matrix_uuid="1234"
                )

                # chekc if entity_id and as_of_date are as index 
                should_be = ['entity_id', 'as_of_date']
                actual_indices = result.index.names

                TestCase().assertListEqual(should_be, actual_indices)

                # last element in the DF should be the label
                last_col = 'booking'
                output = result.columns.values[-1] # label name

                TestCase().assertEqual(last_col, output)

                # number of columns must be the sum of all the columns on each feature table + 1 for the label 
                TestCase().assertEqual(result.shape[1], 4+1, 
                                       "Number of features and label doesn't match")

                # number of rows 
                assert result.shape[0] ==  5
                TestCase().assertEqual(result.shape[0], 5, 
                                       "Number of rows doesn't match")

                # types of the final df should be float32
                types = set(result.apply(lambda x: x.dtype == 'float32').values)
                TestCase().assertTrue(types, "NOT all cols in matrix are float32!")


class TestBuildMatrix(TestCase):
    
    def test_train_matrix(self):
        dates = [
                    datetime.datetime(2016, 1, 1, 0, 0),
                    datetime.datetime(2016, 2, 1, 0, 0),
                    datetime.datetime(2016, 3, 1, 0, 0),
                ]

        features = [["f1", "f2"], ["f3", "f4"]]
        
        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states,
            )

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    experiment_hash=experiment_hash,
                    engine=engine,
                )

                good_metadata = {
                    "matrix_id": "hi",
                    "state": "active",
                    "label_name": "booking",
                    "end_time": datetime.datetime(2016, 3, 1, 0, 0),
                    "feature_start_time": datetime.datetime(2016, 1, 1, 0, 0),
                    "label_timespan": "1 month",
                    "max_training_history": "1 month",
                    "test_duration": "1 month",
                    "indices": ["entity_id", "as_of_date"],
                }

                feature_dictionary = {
                    f"features{i}": feature_list
                    for i, feature_list in enumerate(features)
                }

                uuid = filename_friendly_hash(good_metadata)
                builder.build_matrix(
                    as_of_times=dates,
                    label_name="booking",
                    label_type="binary",
                    feature_dictionary=feature_dictionary,
                    matrix_metadata=good_metadata,
                    matrix_uuid=uuid,
                    matrix_type="train",
                )
            
                assert len(matrix_storage_engine.get_store(uuid).design_matrix) == 5

                #engine_ = create_engine(postgresql.url())
                #assert (
                builder.sessionmaker().query(Matrix)#.get(uuid).feature_dictionary
                #   == feature_dictionary
                #)     
    
    def test_test_matrix(self):
        dates = [
                    datetime.datetime(2016, 1, 1, 0, 0),
                    datetime.datetime(2016, 2, 1, 0, 0),
                    datetime.datetime(2016, 3, 1, 0, 0),
                ]

        features = [["f1", "f2"], ["f3", "f4"]]

        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states,
            )

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    experiment_hash=experiment_hash,
                    engine=engine,
                )

                good_metadata = {
                    "matrix_id": "hi",
                    "state": "active",
                    "label_name": "booking",
                    "end_time": datetime.datetime(2016, 3, 1, 0, 0),
                    "feature_start_time": datetime.datetime(2016, 1, 1, 0, 0),
                    "label_timespan": "1 month",
                    "max_training_history": "1 month",
                    "test_duration": "1 month",
                    "indices": ["entity_id", "as_of_date"],
                }

                feature_dictionary = {
                    f"features{i}": feature_list
                    for i, feature_list in enumerate(features)
                }

                uuid = filename_friendly_hash(good_metadata)
                builder.build_matrix(
                    as_of_times=dates,
                    label_name="booking",
                    label_type="binary",
                    feature_dictionary=feature_dictionary,
                    matrix_metadata=good_metadata,
                    matrix_uuid=uuid,
                    matrix_type="test",
                )

                assert len(matrix_storage_engine.get_store(uuid).design_matrix) == 5
    

    def test_nullcheck(self):
        dates = [
                    datetime.datetime(2016, 1, 1, 0, 0),
                    datetime.datetime(2016, 2, 1, 0, 0),
                    datetime.datetime(2016, 3, 1, 0, 0),
                ]

        features = [["f1", "f2"], ["f3", "f4"]]

        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states,
            )

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    experiment_hash=experiment_hash,
                    engine=engine,
                )

                good_metadata = {
                    "matrix_id": "hi",
                    "state": "active",
                    "label_name": "booking",
                    "end_time": datetime.datetime(2016, 3, 1, 0, 0),
                    "feature_start_time": datetime.datetime(2016, 1, 1, 0, 0),
                    "label_timespan": "1 month",
                    "max_training_history": "1 month",
                    "test_duration": "1 month",
                    "indices": ["entity_id", "as_of_date"],
                }

                feature_dictionary = {
                    f"features{i}": feature_list
                    for i, feature_list in enumerate(features)
                }

                uuid = filename_friendly_hash(good_metadata)
                with self.assertRaises(ValueError):
                    builder.build_matrix(
                        as_of_times=dates,
                        label_name="booking",
                        label_type="binary",
                        feature_dictionary=feature_dictionary,
                        matrix_metadata=good_metadata,
                        matrix_uuid=uuid,
                        matrix_type="other",
                    )
                
    
    def test_replace_false_rerun(self):
        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states,
            )

            dates = [
                datetime.datetime(2016, 1, 1, 0, 0),
                datetime.datetime(2016, 2, 1, 0, 0),
                datetime.datetime(2016, 3, 1, 0, 0),
            ]

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    experiment_hash=experiment_hash,
                    engine=engine,
                    replace=False,
                )

                feature_dictionary = {
                    "features0": ["f1", "f2"],
                    "features1": ["f3", "f4"],
                }
                matrix_metadata = {
                    "matrix_id": "hi",
                    "state": "active",
                    "label_name": "booking",
                    "end_time": datetime.datetime(2016, 3, 1, 0, 0),
                    "feature_start_time": datetime.datetime(2016, 1, 1, 0, 0),
                    "label_timespan": "1 month",
                    "test_duration": "1 month",
                    "indices": ["entity_id", "as_of_date"],
                }
                uuid = filename_friendly_hash(matrix_metadata)
                builder.build_matrix(
                    as_of_times=dates,
                    label_name="booking",
                    label_type="binary",
                    feature_dictionary=feature_dictionary,
                    matrix_metadata=matrix_metadata,
                    matrix_uuid=uuid,
                    matrix_type="test",
                )

                assert len(matrix_storage_engine.get_store(uuid).design_matrix) == 5
                # rerun
                builder.make_entity_date_table = mock.Mock()
                builder.build_matrix(
                    as_of_times=dates,
                    label_name="booking",
                    label_type="binary",
                    feature_dictionary=feature_dictionary,
                    matrix_metadata=matrix_metadata,
                    matrix_uuid=uuid,
                    matrix_type="test",
                )
                assert not builder.make_entity_date_table.called
    
    def test_replace_true_rerun(self):
        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states,
            )
            matrix_metadata = matrix_metadata_creator(
                state="active", test_duration="1month", label_name="booking"
            )

            dates = [
                datetime.datetime(2016, 1, 1, 0, 0),
                datetime.datetime(2016, 2, 1, 0, 0),
                datetime.datetime(2016, 3, 1, 0, 0),
            ]

            feature_dictionary = {"features0": ["f1", "f2"], "features1": ["f3", "f4"]}
            uuid = filename_friendly_hash(matrix_metadata)
            build_args = dict(
                as_of_times=dates,
                label_name="booking",
                label_type="binary",
                feature_dictionary=feature_dictionary,
                matrix_metadata=matrix_metadata,
                matrix_uuid=uuid,
                matrix_type="test",
            )

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    experiment_hash=experiment_hash,
                    engine=engine,
                    replace=True,
                )

                builder.build_matrix(**build_args)

                assert len(matrix_storage_engine.get_store(uuid).design_matrix) == 5
                assert builder.sessionmaker().query(Matrix).get(uuid)
                # rerun
                builder.build_matrix(**build_args)
                assert len(matrix_storage_engine.get_store(uuid).design_matrix) == 5
                assert builder.sessionmaker().query(Matrix).get(uuid)
    


import logging
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import itertools
from IPython.display import display


from descriptors import cachedproperty
from scipy.stats import spearmanr

from triage.component.postmodeling.model_analyzer import ModelAnalyzer
from triage.component.postmodeling.error_analysis import generate_error_analysis, output_all_analysis


class PostmodelingAnalyzer: 

    def __init__(self, engine, model_groups, experiment_hashes, project_path=None, train_end_times=None, use_all_model_groups=False) -> None:
        self.model_groups = model_groups
        self.experiment_hashes = experiment_hashes # TODO made experiment hashes into a list to plot models from different experiments for MVESC, there's probably a better way to generalize this
        self.engine = engine
        self.project_path = project_path
        if use_all_model_groups: # shortcut to listing out all model groups for an experiment
            self.use_all_model_groups()
        self.models = self.get_model_ids()

        #TODO: Would be good to let the user restrict the train_end_times included in the report

        
    @property
    def model_ids(self):
        model_ids = list()
        for mg, d in self.models.items():
            for t, mod_obj in d.items():
                model_ids.append(mod_obj.model_id)

    @property
    def model_types(self):
        pass

    def use_all_model_groups(self):
        experiment_hashes = "','".join(self.experiment_hashes)
        q = f"""
                select model_group_id 
                from triage_metadata.models
                    join triage_metadata.experiment_models using (model_hash)
                where experiment_hash in ('{experiment_hashes}')
            """
        model_groups = pd.read_sql(q, self.engine)
        self.model_groups = model_groups['model_group_id'].to_list()

    # TODO: revise this to show the only the model_group_id, list of model ids, model type, and hyperparameters
    def display_model_groups(self):
        data_dict = []
        for mg in self.model_groups:
            for train_end_time in self.models[mg]:
                model_analyzer = self.models[mg][train_end_time]
                data_dict.append([mg, train_end_time, model_analyzer.model_id, model_analyzer.model_type, model_analyzer.hyperparameters])
        all_models = pd.DataFrame(data_dict, columns=['model_group_id', 'train_end_time', 'model_id', 'model_type', 'hyperparameters'])

        # displaying the model_group_ids, model_type, and the hyperparameters
        to_print = all_models.groupby('model_group_id').nth(1)[['model_type', 'hyperparameters']].reset_index().to_dict(orient='records')

        for m in to_print:
            print(m)

        return all_models
    
    def print_model_summary(self):
        ''' This is mostly to be used as a key for modeling report plots (just as a model group number to model group mapping)'''
        data_dict = []
        for mg in self.model_groups:
            for train_end_time in self.models[mg]:
                model_analyzer = self.models[mg][train_end_time]
                data_dict.append([mg, train_end_time, model_analyzer.model_id, model_analyzer.model_type, model_analyzer.hyperparameters])
        
        all_models = pd.DataFrame(data_dict, columns=['model_group_id', 'train_end_time', 'model_id', 'model_type', 'hyperparameters'])
        # to_print = all_models.groupby('model_group_id').nth(1)[['model_type', 'hyperparameters']].reset_index().to_dict(orient='records')

        for i, model in all_models.groupby('model_group_id').nth(1)[['model_type', 'hyperparameters']].reset_index().iterrows():
            print(f"{model['model_group_id']} - {model['model_type']} with ({model['hyperparameters']}) ")

    def cohort_summary(self):
        q = f"""
            select distinct on(train_end_time)
                -- matrix_uuid,
                evaluation_start_time as train_end_time,
                num_labeled_examples as cohort_size,
                num_positive_labels,
                case when num_labeled_examples > 0 then num_positive_labels::float/num_labeled_examples else 0 end as label_base_rate
            from triage_metadata.experiment_matrices join test_results.evaluations using(matrix_uuid)
            where experiment_hash in ('{"','".join(self.experiment_hashes)}') and subset_hash = ''
            order by 1
        """

        matrices = pd.read_sql(q, self.engine)

        print(matrices)
    

    def subset_summary(self, subset_hash):
        q = f"""
            select distinct on(train_end_time)
                -- matrix_uuid,
                evaluation_start_time as train_end_time,
                num_labeled_examples as cohort_size,
                num_positive_labels,
                case when num_labeled_examples > 0 then num_positive_labels::float/num_labeled_examples else 0 end as label_base_rate
            from triage_metadata.experiment_matrices join test_results.evaluations using(matrix_uuid)
            where experiment_hash in ('{"','".join(self.experiment_hashes)}') 
                and subset_hash = '{subset_hash}'
            order by 1
        """

        matrices = pd.read_sql(q, self.engine)

        print(matrices)

    def plot_model_group_performance(self, metric, parameter):
        pass        

        # q = """
        #     select
        
        # """

    def get_model_ids(self):
        """ For the model group ids, fetch the model_ids and initialize the datastructure

            The data structure is a dictionary of dictionaries that maps the individual models of groups to their ModelAnalyzer class
                {model_group_id: {train_end_time: ModelAnalyzer(model_id)}}
                Here, the train_end_time is saved as a string of with format 'YYYY-MM-DD'
        """

        model_groups = "', '".join([str(x) for x in self.model_groups])
        experiment_hashes = "', '".join(self.experiment_hashes)
        q = f"""
            select distinct on (model_group_id, train_end_time)
                model_id, 
                train_end_time::date,
                model_group_id
            from triage_metadata.models 
                join triage_metadata.experiment_models using(model_hash)
            where experiment_hash in ('{experiment_hashes}')
            and model_group_id in ('{model_groups}')        
            """  
        # TODO do we really need experiment_hashes here? can we query with only model_group_ids?

        # TODO: modify to remove pandas
        models = pd.read_sql(q, self.engine).to_dict(orient='records')

        d = dict()
        for experiment_hash in self.experiment_hashes:
            q = f"""
                select distinct on (model_group_id, train_end_time)
                    model_id, 
                    to_char(train_end_time::date, 'YYYY-MM-DD') as train_end_time,
                    model_group_id
                from triage_metadata.models 
                    join triage_metadata.experiment_models using(model_hash)
                where experiment_hash='{experiment_hash}'
                and model_group_id in ('{model_groups}')        
                """  

            # TODO: modify to remove pandas
            models = pd.read_sql(q, self.engine).to_dict(orient='records')

            for m in models:
                if m['model_group_id'] in d:
                    d[m['model_group_id']][m['train_end_time']] = ModelAnalyzer(m['model_id'], self.engine)
                else:
                    d[m['model_group_id']] = {m['train_end_time']: ModelAnalyzer(m['model_id'], self.engine)}

        return d 
    
    def _get_subplots(self, subplot_width=3, subplot_len=None, n_rows=None, n_cols=None, sharey=False, sharex=False):
        """"""

        if subplot_len is None:
            subplot_len = subplot_width

        # num of rows == number of train_end_times
        if n_rows is None: 
            n_rows = len(self.models[self.model_groups[0]])
        
        # num of cols == number of model groups
        if n_cols is None: 
            n_cols = len(self.model_groups)
        
        fig, axes = plt.subplots(
            n_rows,
            n_cols,
            figsize = (subplot_width*n_cols, subplot_len*n_rows),
            dpi=100,
            sharey=sharey, 
            sharex=sharex
        )

        return fig, axes


    def _make_plot_grid(self, plot_type, subplot_width=3, subplot_len=None, sharey=False, sharex=False, **kw):
        """
            Abstracts out generating the plot grid (time x model group) for comparing
            model_scores,  
        """
        fig, axes = self._get_subplots(subplot_width=subplot_width, subplot_len=subplot_len, sharey=sharey, sharex=sharex)
        
        print(len(axes), len(axes[0]))

        for j, mg in enumerate(self.models):
            for i, train_end_time in enumerate(self.models[mg]):
                model_analyzer = self.models[mg][train_end_time]

                plot_func = getattr(model_analyzer, plot_type)

                if (len(self.models) == 1) or (len(self.models[mg]) == 1):
                    if (len(self.models) == 1) and (len(self.models[mg]) == 1):
                        ax = axes
                    else:
                        ax = axes[i]
                else:
                    ax = axes[i, j]
                ax = plot_func(ax=ax, **kw)

                if j==0:
                    ax.set_ylabel(f'{train_end_time}')
                else:
                    ax.set_ylabel('')
                
                ax.set_xlabel('')

        fig.tight_layout()

        return fig

    def plot_score_distributions(self, use_labels=False):
        """for the model group ids plot score grid"""

        if use_labels:
            self._make_plot_grid(plot_type='plot_score_label_distribution')
        else:
            self._make_plot_grid(plot_type='plot_score_distribution')
        
    def plot_calibration_curves(self):
        """calibration curves for all models"""
        self._make_plot_grid(plot_type='plot_calibration_curve')

    def plot_prk_curves(self, **kw):
        self._make_plot_grid(plot_type='plot_precision_recall_curve', **kw)
        
    def plot_recall_curves_overlaid(self, n_splits=None, **kw,):
        # Number of columns    
        if n_splits is None:
            n_cols = len(self.models[self.model_groups[0]])
        else:
            n_cols = n_splits
            
        fig, axes = self._get_subplots(
            subplot_width=3, 
            subplot_len=3, 
            sharey=True, 
            sharex=True, 
            n_cols=n_cols, 
            n_rows=1
        )
        
        as_of_dates = list(sorted(self.models[self.model_groups[0]].keys()))[-n_cols:]
        
        for i, aod in enumerate(as_of_dates):
            for model_group in sorted(self.model_groups):
                mod = self.models[model_group].get(aod)
                if mod is None:
                    continue

                mod.plot_precision_recall_curve(only_recall=True,ax = axes[i], title_string=aod)
                axes[i].set(alpha=0.1)
                axes[i].legend().remove()
                axes[i].set_xlabel('Population pct (k%)')
                axes[i].set_ylabel('recall@k')
                sns.despine()
        plt.tight_layout()
        axes[-1].legend(sorted(self.model_groups))
        
        
        

  
    def plot_bias_threshold(self, attribute_name, attribute_values, bias_metric):
        """
            Plot bias_metric for the specified list of attribute_values for a particular attribute_name across different thresholds (list %)
        """
        fig, axes = self._get_subplots(subplot_width=6, n_rows=len(attribute_values), n_cols=len(self.models[self.model_groups[0]]))
        for _, mg in enumerate(self.models):
            for i, attribute_value in enumerate(attribute_values):
                for j, train_end_time in enumerate(self.models[mg]):
                    mode_analyzer = self.models[mg][train_end_time]

                    mode_analyzer.plot_bias_threshold_curve(
                        attribute_name=attribute_name,
                        attribute_value=attribute_value,
                        bias_metric=bias_metric,
                        ax=axes[i, j]
                    )
                    if j==0:
                        axes[i, j].set_ylabel(f'{attribute_name}:{attribute_value}')
                    else:
                        axes[i, j].set_ylabel('')
                    if i == 0:
                        axes[i, j].set_title(f'{train_end_time}')
                    else:
                        axes[i, j].set_title('')
        fig.suptitle(f"{bias_metric} Threshold Curve for {attribute_name}")
        fig.tight_layout()

    def plot_precision_threshold(self):
        """
            Plot precision against threshold (list %)
        """
        if len(self.models) <= 1:
            print("Not available when there is only one model group (look at plot_prk_curves instead)")
            return
        fig, axes = self._get_subplots(subplot_width=6, n_cols=1, sharey=True)
        for _, mg in enumerate(self.models):
            for j, train_end_time in enumerate(self.models[mg]):
                mode_analyzer = self.models[mg][train_end_time]
                mode_analyzer.plot_precision_threshold_curve(
                    ax=axes[j]
                )
                axes[j].set_title(f'{train_end_time}')
        fig.suptitle("Precision Threshold Curve")
        fig.tight_layout()

    def plot_feature_importance(self, n_top_features=20):
        """ plot all feature importance  """
        self._make_plot_grid(plot_type='plot_feature_importance', subplot_width=7, n_top_features=n_top_features)
        
            
    def plot_feature_group_importance(self, n_top_groups=20):
        """ plot all feature group importance  """
        self._make_plot_grid(plot_type='plot_feature_group_importance', subplot_width=7, n_top_groups=n_top_groups)


    def create_scatter_disparity_performance(self, metric, parameter, aeq_parameter, attr_col, attribute_values, 
                                         performance_col='stochastic_value', bias_metric='tpr', tiebreaker='worst', flip_disparity=False, 
                                         mitigated_tags=[], mitigated_bdfs=[], mitigated_performances=[], ylim=None):
        """
            Create scatterplot of one bias metric (e.g. tpr disparity) vs an evaluation metric (e.g. precision at some threshold) for a particular list of attribute value (must be from the same attribute)
            A simplified version of the scatterplot function here: https://github.com/dssg/fairness_tutorial/blob/master/notebooks/bias_reduction_with_outputs.ipynb 
        """
        # TODO add legend for model identification?
        evals_df_list = {}
        aequitas_df_list = {}
        fig, axes = self._get_subplots(subplot_width=6, n_rows=len(attribute_values))
        
        for k, attribute_value in enumerate(attribute_values):
            for i, mg in enumerate(self.models):
                for j, train_end_time in enumerate(self.models[mg]):
                    mode_analyzer = self.models[mg][train_end_time]
                    if i == 0:
                        evals_df_list[train_end_time] = []
                        aequitas_df_list[train_end_time] = []
                    evals_df_list[train_end_time].append(mode_analyzer.get_evaluations(metrics={metric: [parameter]})) # TODO only allow one eval metric here?
                    aequitas_df_list[train_end_time].append(mode_analyzer.get_aequitas())

            for j, train_end_time in enumerate(self.models[self.model_groups[0]]):

                evals_df = pd.concat(evals_df_list[train_end_time])
                aequitas_df = pd.concat(aequitas_df_list[train_end_time])
                # filter aequitas by eval metric
                aequitas_df = aequitas_df.loc[(aequitas_df['parameter']==aeq_parameter) * (aequitas_df['tie_breaker']==tiebreaker)]
                disparity_df = aequitas_df.loc[(aequitas_df['attribute_name']==attr_col) & (aequitas_df['attribute_value']==attribute_value)].copy()
                disparity_metric = bias_metric + '_disparity'
                scatter_schema = ['model_id', performance_col, 'attribute_name', 'attribute_value', bias_metric, disparity_metric, 'model_tag']
                if flip_disparity:
                    disparity_df[disparity_metric]= disparity_df.apply(lambda x: 1/x[disparity_metric] , axis=1)
                scatter = pd.merge(evals_df, disparity_df, how='left', on=['model_id'], sort=True, copy=True)
                scatter = scatter[['model_id', performance_col, 'attribute_name', 'attribute_value', bias_metric, disparity_metric]].copy()
                scatter['model_tag'] = 'Other Models'
                scatter.sort_values('stochastic_value', ascending = False, inplace=True, ignore_index=True)
                scatter_final = pd.DataFrame()

                ax = axes[k,j]
                ax.scatter(
                    x='stochastic_value', y=disparity_metric,
                    data=scatter

                )
                if not scatter_final.empty:
                    ax.scatter(
                        x='stochastic_value', y=disparity_metric,
                        data=scatter_final
                    )
                
                if j==0:
                    axes[k, j].set_ylabel(f'{attr_col}:{attribute_value}')
                else:
                    axes[k, j].set_ylabel('')
                if k == 0:
                    axes[k, j].set_title(f'{train_end_time}')
                else:
                    axes[k, j].set_title('')
                axes[k, j].set_xlabel(f'{metric}{parameter}')

                if ylim:
                    plt.ylim(0, 10)
        flip_placeholder = 'Flipped' if flip_disparity else ''
        fig.suptitle(f'{flip_placeholder} {disparity_metric} vs. {metric}{parameter} for {attr_col}')


    def calculate_crosstabs_pos_vs_neg(self, project_path, thresholds, table_name='crosstabs', **kwargs):
        """ Generate crosstabs for the predicted positives (top-k) vs the rest

        args:
            project_path (str): Path where the experiment artifacts (models and matrices) are stored
            thresholds (Dict{str: Union[float, int}]): A dictionary that maps threhold type to the threshold
                                                    The threshold type can be one of the rank columns in the test_results.predictions_table
            table_name (str, optional): Table name to use in the db's `test_results` schema. Defaults to crosstabs.
                                        If the table exists, results are appended
            **kwargs: This method can take other arguments sent to ModelAnalyzer.crosstabs_pos_vs_ng function
        """

        for i, mg in enumerate(self.models):
            for j, train_end_time in enumerate(self.models[mg]):
                model_analyzer = self.models[mg][train_end_time]

                df = model_analyzer.crosstabs_pos_vs_neg(
                    project_path=project_path,
                    thresholds=thresholds,
                    table_name=table_name,
                    **kwargs
                )


    def display_crosstab_pos_vs_neg(
        self, 
        threshold_type,
        threshold, 
        table_name='crosstabs', 
        project_path=None,
        display_n_features=40,
        filter_features=None,
        support_threshold=0.1,
        show_plots=True,
        return_dfs=True):

        """ display crosstabs for one threshold for all the models in the model groups
            
        Args:
            threshold_type (str): Type of rank threshold to use in splitting predicted positives from negatives. 
                                Has to be one of the rank columns in the test_results.predictions_table

            threshold (Union[int, float]): The rank threshold of the specified type. If the threshold type is an absolute, integer. If percentage, should be a float between 0 and 1
                        
            table_name (str, optional): Table name to fetch from/write to in the db's `test_results` schema. Defaults to crosstabs.
            
            project_path (str, optional): Path where the experiment artifacts are stored. Optional if the crosstabs are already calculated,
                required if the crosstab need to be calculated
            
            display_n_features (int, optional): Number of features to return. defaults to 40 (sorted by mean ratio desc). This is ignored if `filter_features` is specified

            filter_features (List[str], optional): The list of features that we are interested in. If not specified, `display_n_features` features are returned

            support_threshold (float, optional): The threshold of pct support for the feature (instances with non-zero values) among predicted positives 
        """

        dfs = dict()

        for i, mg in enumerate(self.models):
            for j, train_end_time in enumerate(self.models[mg]):
                model_analyzer = self.models[mg][train_end_time]

                idx = i * len(self.models[self.model_groups[0]]) + j 


                if len(self.models) == 1 and len(self.models[mg]) == 1:
                    axes = [axes] 
                try:
                    df = model_analyzer.display_crosstabs_pos_vs_neg(
                        threshold_type=threshold_type,
                        threshold=threshold,
                        display_n_features=display_n_features,
                        filter_features=filter_features,
                        support_threshold=support_threshold,
                        table_name=table_name,
                        # ax=(axes if show_plots else None),
                        show_plot=False, #TODO -- remove this parameter from 
                        return_df=True
                    )

                    dfs[model_analyzer.model_id] = df
                except ValueError as e:
                    logging.error('Please run calculate_crosstabs_pos_vs_neg function to calculate crosstabs first for all models!')
                    raise e
                
                print(f'\nModel Group: {mg}, Validation date: {train_end_time}'.center(30, ' '))
                display(df)

        if return_dfs:
            return dfs
        
    def _pairwise_feature_importance_comparison_single_split(self, train_end_time, n_top_features, model_group_ids=None,plot=True):
        """ For a given train_end_time, compares the top n features (highest absolute importance) of two models 
            
            Args:
                train_end_time (str): The prediction date we care about in YYYY-MM-DD format
                n_top_features (int): Number of features to consider for the comparispn
                model_group_ids (int, optional): Model group ids to consider, if not provided, all model groups included in the report are used
                plot (bool, optional): Whether to plot the results. Defaults to True.
        """
        
        feature_lists = dict()
        
        if model_group_ids is not None:
            
            for mg in model_group_ids:
                feature_lists[mg] = self.models[mg][train_end_time].get_feature_importances(n_top_features=n_top_features) 
                if feature_lists[mg].empty:
                    logging.warning('No feature importance values were found for model group {mg}. Excluding from comparison')
                    feature_lists.pop(mg)
        # By default all feature importance values are considered
        else:
            model_group_ids = self.model_groups
            for mg, m in self.models.items():
                feature_lists[mg] = m[train_end_time].get_feature_importances(n_top_features=n_top_features) 
                if feature_lists[mg].empty:
                    logging.warning('No feature importance values were found for model group {mg}. Excluding from comparison')
                    feature_lists.pop(mg)
            
        pairs = list(itertools.combinations(feature_lists.keys(), 2))
        
        logging.info(f'Performing {len(pairs)} comparisons')
        
        metrics = ['jaccard', 'overlap', 'rank_corr']
        results = dict()
        
        for m in metrics:
            results[m] = pd.DataFrame(index=sorted(model_group_ids), columns=sorted(model_group_ids))
            # filling the diagonal with 1
            results[m].values[[np.arange(results[m].shape[0])]*2] = 1
            
        for model_group_pair in pairs:
            model_group_pair = sorted(model_group_pair)
            logging.info(f'Comparing {model_group_pair[0]} and {model_group_pair[1]}')

            df1 = feature_lists[model_group_pair[0]]
            df2 = feature_lists[model_group_pair[1]]
            
            f1 = set(df1.feature)
            f2 = set(df2.feature)
            
            if (len(f1) == 0 or len(f2)) == 0:
                logging.error('No feature importance available for the models!') 
                continue

            inter = f1.intersection(f2)
            un = f1.union(f2)    
            results['jaccard'].loc[model_group_pair[1], model_group_pair[0]] = len(inter)/len(un)

            # If the list sizes are not equal, using the smallest list size to calculate simple overlap
            results['overlap'].loc[model_group_pair[1], model_group_pair[0]] = len(inter)/ min(len(f1), len(f2))

            # calculating rank correlation
            df1.sort_values('feature_importance', ascending=False, inplace=True)
            df2.sort_values('feature_importance', ascending=False, inplace=True)

            # only returning the corr coefficient, not the p-value
            results['rank_corr'].loc[model_group_pair[0], model_group_pair[1]] = spearmanr(df1.feature.iloc[:], df2.feature.iloc[:])[0]

    
        if plot:
            fig, axes = plt.subplots(1, len(metrics), figsize=(10, 3))            
            
            for i, m in enumerate(metrics):
                sns.heatmap(
                    data=results[m].fillna(0),
                    cmap='Greens',
                    vmin=0,
                    vmax=1,
                    annot=True,
                    linewidth=0.1,
                    ax=axes[i]
                )

                axes[i].set_title(m)

            fig.suptitle(train_end_time)
            fig.tight_layout()
        
        return results
    

    def _pairwise_list_comparison_single_fold(self, threshold_type, threshold, train_end_time, matrix_uuid=None, plot=True):
        """For a given train_end_time, compares the lists generated by the analyzed model groups

        Args:
            threshold_type (str): Type of the ranking to use. Has to be one of the four ranking types used in triage
                    - rank_pct_no_ties 
                    - rank_pct_with_ties
                    - rank_abs_no_ties
                    - rank_abs_with_ties
            threshold (Union[float, int]): The threshold rank for creating the list. Int for 'rank_abs_*' and Float for 'rank_pct_*'
            train_end_time (str): The prediction date we care about in YYYY-MM-DD format
        """

        lists = dict()

        for mg, m in self.models.items():
            lists[mg] = m[train_end_time].get_top_k(threshold_type, threshold, matrix_uuid)

        
        pairs = list(itertools.combinations(lists.keys(), 2))

        logging.info(f'Performing {len(pairs)} comparisons')

        metrics = ['jaccard', 'overlap', 'rank_corr']
        results = dict()

        # Initializing three data frames to hold pairwise metrics
        for m in metrics:
            results[m] = pd.DataFrame(index=sorted(self.model_groups), columns=sorted(self.model_groups))
            results[m].values[[np.arange(results[m].shape[0])]*2] = 1

        for model_group_pair in pairs:
            logging.info(f'Comparing {model_group_pair[0]} and {model_group_pair[1]}')

            model_group_pair = sorted(model_group_pair)
            
            df1 = lists[model_group_pair[0]]
            df2 = lists[model_group_pair[1]]

            # calculating jaccard similarity and overlap
            entities_1 = set(df1.entity_id)
            entities_2 = set(df2.entity_id)

            if (len(entities_1) == 0 or len(entities_2)) == 0:
                logging.error('No prediction saved for the models!') 

            inter = entities_1.intersection(entities_2)
            un = entities_1.union(entities_2)    
            results['jaccard'].loc[model_group_pair[1], model_group_pair[0]] = len(inter)/len(un)

            # If the list sizes are not equal, using the smallest list size to calculate simple overlap
            results['overlap'].loc[model_group_pair[1], model_group_pair[0]] = len(inter)/ min(len(entities_1), len(entities_2))

            # calculating rank correlation
            df1.sort_values('score', ascending=False, inplace=True)
            df2.sort_values('score', ascending=False, inplace=True)

            # only returning the corr coefficient, not the p-value
            if len(df1) == len(df2):
                results['rank_corr'].loc[model_group_pair[0], model_group_pair[1]] = spearmanr(df1.entity_id.iloc[:], df2.entity_id.iloc[:])[0]
            else:
                logging.warning(f'Not calculating rank correlation. List sizes are not equal ({len(df1)}, {len(df2)})')

        if plot:
            fig, axes = plt.subplots(1, len(metrics), figsize=(10, 3))            
            
            for i, m in enumerate(metrics):
                sns.heatmap(
                    data=results[m].fillna(0),
                    cmap='Greens',
                    vmin=0,
                    vmax=1,
                    annot=True,
                    linewidth=0.1,
                    ax=axes[i]
                )

                axes[i].set_title(m)

            fig.suptitle(train_end_time)
            fig.tight_layout()
        return results

    def _get_individual_model_ids(self):
        """Get individual model ids associated with the given model groups
        """
        model_groups = ", ".join([str(x) for x in self.model_groups])

        q = f"""
            select
                model_id 
            from triage_metadata.models
            where model_group_id in({model_groups})
        """
        df_model_ids = pd.read_sql(q, self.engine)

        return df_model_ids.model_id.to_list()


    def execute_error_analysis(self, model_ids=None):
        """Generates the error analysis of a model
            args:
                model_ids (int, optional): A list of model_ids we are interested in analyzing. If not provided, all models are considered.
        """
        if model_ids is None:
            model_ids = self.model_ids

        for model_id in model_ids:
            logging.info(f"generating error analysis for model id {model_id}")
            results = generate_error_analysis(model_id, self.engine, self.project_path)
            output_all_analysis(results)
    
    
    def pairwise_top_k_list_comparison(self, threshold_type, threshold, train_end_times=None, matrix_uuid=None, plot=True):
        """
            Compare the top-k lists for the given train_end_times for all model groups considered (pairwise)
            
            Args:
                threshold_type (str): Type of the ranking to use. Has to be one of the four ranking types used in triage
                        - rank_pct_no_ties 
                        - rank_pct_with_ties
                        - rank_abs_no_ties
                        - rank_abs_with_ties
                threshold (Union[float, int]): The threshold rank for creating the list. Int for 'rank_abs_*' and Float for 'rank_pct_*'
                train_end_times (Optional, List[str]): The prediction date we care about in YYYY-MM-DD format
        """

        # If no train_end_times are provided, we consider all the train_end_times
        # NOTE -- Assuming that the all model groups have the same train_end_times
        if train_end_times is None:
            train_end_times = self.models[self.model_groups[0]].keys()

        for train_end_time in train_end_times:
            self._pairwise_list_comparison_single_fold(
                threshold=threshold,
                threshold_type=threshold_type,
                train_end_time=train_end_time,
                matrix_uuid=matrix_uuid,
                plot=plot
            )
            
    def pairwise_feature_importance_comparison(self, n_top_features, model_groups=None, train_end_times=None, plot=True):
        """
            Compare the top-k lists for the given train_end_times for all model groups considered (pairwise)
            
            Args:
                threshold_type (str): Type of the ranking to use. Has to be one of the four ranking types used in triage
                        - rank_pct_no_ties 
                        - rank_pct_with_ties
                        - rank_abs_no_ties
                        - rank_abs_with_ties
                threshold (Union[float, int]): The threshold rank for creating the list. Int for 'rank_abs_*' and Float for 'rank_pct_*'
                train_end_times (Optional, List[str]): The prediction date we care about in YYYY-MM-DD format
        """

        # If no train_end_times are provided, we consider all the train_end_times
        # NOTE -- Assuming that the all model groups have the same train_end_times
        if train_end_times is None:
            train_end_times = self.models[self.model_groups[0]].keys()
            

        for train_end_time in train_end_times:
            self._pairwise_feature_importance_comparison_single_split(
                train_end_time=train_end_time,
                n_top_features=n_top_features,
                model_group_ids=model_groups,
                plot=plot
            )


    


import numpy as np
import pandas as pd
import pandas.api.types as ptypes

import pytest
from unittest import TestCase
from numpy.testing import assert_array_equal

from triage.component.catwalk.baselines.rankers import PercentileRankOneFeature
from triage.component.catwalk.baselines.rankers import BaselineRankMultiFeature
from triage.component.catwalk.baselines.rankers import LinearRanker
from triage.component.catwalk.baselines.thresholders import SimpleThresholder
from triage.component.catwalk.baselines.thresholders import get_operator_method
from triage.component.catwalk.baselines.thresholders import OPERATOR_METHODS
from triage.component.catwalk.exceptions import BaselineFeatureNotInMatrix


@pytest.fixture(scope="class")
def data(request):
    X_train = pd.DataFrame( 
        {
            "x1": [0, 1, 2, 56, 25, 8, -3, 89],
            "x2": [0, 23, 1, 6, 5, 3, 18, 7],
            "x3": [1, 12, 5, -6, 2, 5, 3, -3],
            "x4": [6, 13, 4, 5, 35, 6, 43, 74],
            "x6": [400, 400, 400, 300, 300, 300, 300, 300],
            "x7": [1, 12, 12, -6, -6, 5, 5, -3],
        }
    )
    y_train = [0, 1, 0, 1, 1, 1, 3, 0]
    X_test = pd.DataFrame(
        {
            "x1": [4, 14, 0, 6, 25, 8, -3, 4],
            "x2": [6, -1, 1, 24, 5, 3, 18, 39],
            "x3": [1, 7, 4, 57, 2, 5, 3, 2],
            "x4": [7, 3, 6, 39, 35, 6, 43, -6],
            "x6": [305, 305, 305, 305, 401, 401, 401, 401],
            "x7": [1, 7, 7, 57, 2, 5, 3, 2],
        }
    )
    y_test = [1, 3, 0, 0, 0, 0, 0, 1]

    request.cls.data = {
        "X_train": X_train,
        "X_test": X_test,
        "y_train": y_train,
        "y_test": y_test,
    }

@pytest.fixture(scope="class")
def rules(request):
    request.cls.rules = ["x1 > 0", "x2 <= 1"]

@pytest.fixture(scope="class")
def features(request):
    request.cls.features = ['x1', 'x2', 'x3', 'x4', 'x6', 'x7']

@pytest.fixture(scope="class")
def weights(request):
    request.cls.weights = [0.6, 0.5, 0.4, 0.7, 0.6, 0.3]


def predict_proba_deprecated(ranker, x):
        """ Generate the rank scores and return these.
        """
        # reduce x to the selected set of features
        x = x[ranker.all_feature_names].reset_index(drop=True)

        x = x.sort_values(ranker.all_feature_names, ascending=ranker.all_sort_directions)

        # initialize curr_rank to -1 so the first record will have rank 0 (hence "score"
        # will range from 0 to 1)
        ranks = []
        curr_rank = -1
        prev = []

        # calculate ranks over sorted records, giving ties the same rank
        for rec in x.values:
            if not np.array_equal(prev, rec):
                curr_rank += 1
            ranks.append(curr_rank)
            prev = rec

        # normalize to 0 to 1 range
        x['score'] = [r/max(ranks) for r in ranks]

        # reset back to original sort order, calculate "score" for "0 class"
        scores_1 = x.sort_index()['score'].values
        scores_0 = np.array([1-s for s in scores_1])

        return np.array([scores_0, scores_1]).transpose()

def scores_align_with_ranks(expected_ranks, returned_scores):
    '''
    Helper function to check that scores align with ranks
    correctly for the ranking baselines (e.g., higher ranks
    get higher scores and ties have the same score)
    '''
    df = pd.DataFrame({
        'rank': expected_ranks,
        'score': returned_scores
        }).sort_values('rank', ascending=True)

    curr_rank = None
    curr_score = None

    # Loop through the sorted records to check for any inconsistencies
    for ix, rec in df.iterrows():
        if curr_rank is None:
            curr_rank = rec['rank']
            curr_score = rec['score']
            continue

        if rec['rank'] < curr_rank:
            return RuntimeError('Something has gone wrong with df.sort_values!')
        elif rec['rank'] == curr_rank and rec['score'] != curr_score:
            return False
        elif rec['rank'] > curr_rank and rec['score'] <= curr_score:
            return False

        curr_rank = rec['rank']
        curr_score = rec['score']

    # If we got through the loop without any issues, return True
    return True


def test_scores_align_with_ranks():
    # correct, no ties
    assert scores_align_with_ranks([1,2,3], [0,0.5,1.0])
    # correct, with ties
    assert scores_align_with_ranks([1,2,2,3], [0,0.5,0.5,1.0])
    # incorrect, no ties
    assert not scores_align_with_ranks([1,2,3], [1.0,0.5,0.8])
    # ties with different scores
    assert not scores_align_with_ranks([1,2,2,3], [0,0.5,0.7,1.0])



@pytest.mark.usefixtures("data")
class TestRankOneFeature(TestCase):

    def test_fit(self):
        ranker = PercentileRankOneFeature(feature="x3")
        assert ranker.feature_importances_ is None
        ranker.fit(x=self.data["X_train"], y=self.data["y_train"])
        np.testing.assert_array_equal(
            ranker.feature_importances_, np.array([0, 0, 1, 0, 0, 0])
        )

    def test_ranking_on_unavailable_feature_raises_error(self):
        ranker = PercentileRankOneFeature(feature="x5")
        with self.assertRaises(BaselineFeatureNotInMatrix):
            ranker.fit(x=self.data["X_train"], y=self.data["y_train"])

    def test_predict_proba(self):
        for direction_value in [True, False]:
            ranker = PercentileRankOneFeature(feature="x3", low_value_high_score=direction_value)
            ranker.fit(x=self.data["X_train"], y=self.data["y_train"])
            results = ranker.predict_proba(self.data["X_test"])
            if direction_value:
                expected_ranks = [6, 1, 3, 0, 5, 2, 4, 5]
            else:
               expected_ranks = [0, 5, 3, 6, 1, 4, 2, 1]

            assert scores_align_with_ranks(expected_ranks, results[:,1])


@pytest.mark.usefixtures("data")
class TestRankMultiFeature(TestCase):
    def test_fit(self):
        rules = {'feature': 'x3', 'low_value_high_score': False}
        ranker = BaselineRankMultiFeature(rules=rules)
        assert ranker.feature_importances_ is None
        ranker.fit(x=self.data["X_train"], y=self.data["y_train"])
        np.testing.assert_array_equal(
            ranker.feature_importances_, np.array([0, 0, 1, 0, 0, 0])
        )

    def test_ranking_on_unavailable_feature_raises_error(self):
        rules = [{'feature': 'x5', 'low_value_high_score': False}]
        ranker = BaselineRankMultiFeature(rules=rules)
        with self.assertRaises(BaselineFeatureNotInMatrix):
            ranker.fit(x=self.data["X_train"], y=self.data["y_train"])

    def test_predict_proba_one_feature(self):
        for direction_value in [True, False]:
            rules = {'feature': 'x3', 'low_value_high_score': direction_value}
            ranker = BaselineRankMultiFeature(rules=rules)
            ranker.fit(x=self.data["X_train"], y=self.data["y_train"])
            results = ranker.predict_proba(self.data["X_test"])
            if direction_value:
                expected_ranks = [6, 1, 3, 0, 5, 2, 4, 5]
            else:
                expected_ranks = [0, 5, 3, 6, 1, 4, 2, 1]

            assert scores_align_with_ranks(expected_ranks, results[:,1])

    def test_predict_proba_multi_feature(self):
        rules = [
            {'feature': 'x3', 'low_value_high_score': True},
            {'feature': 'x2', 'low_value_high_score': False}
        ]

        ranker = BaselineRankMultiFeature(rules=rules)
        ranker.fit(x=self.data["X_train"], y=self.data["y_train"])
        results = ranker.predict_proba(self.data["X_test"])

        expected_ranks = [7, 1, 3, 0, 5, 2, 4, 6]

        assert scores_align_with_ranks(expected_ranks, results[:,1])

    def test_predict_proba_no_ties(self):
        for direction_value in [True, False]:
            rules = [{'feature': 'x2', 'low_value_high_score': direction_value}]

            ranker = BaselineRankMultiFeature(rules=rules)
            ranker.fit(x=self.data["X_train"], y=self.data["y_train"])
            results_new = ranker.predict_proba(self.data["X_test"])
            results_deprecated = predict_proba_deprecated(ranker, self.data["X_test"])
            
            assert_array_equal(results_new, results_deprecated)

    
    def test_predict_proba_half_ties(self):
        for direction_value in [True, False]: 
            rules = [{'feature': 'x6', 'low_value_high_score': direction_value}]

            ranker = BaselineRankMultiFeature(rules=rules)
            ranker.fit(x=self.data["X_train"], y=self.data["y_train"])
            results_new = ranker.predict_proba(self.data["X_test"])
            results_deprecated = predict_proba_deprecated(ranker, self.data["X_test"])

            assert_array_equal(results_new, results_deprecated)


    def test_predict_proba_some_ties(self):
        for direction_value in [True, False]: 
            rules = [{'feature': 'x7', 'low_value_high_score': direction_value}]

            ranker = BaselineRankMultiFeature(rules=rules)
            ranker.fit(x=self.data["X_train"], y=self.data["y_train"])
            results_new = ranker.predict_proba(self.data["X_test"])
            results_deprecated = predict_proba_deprecated(ranker, self.data["X_test"])

            assert_array_equal(results_new, results_deprecated)


@pytest.mark.parametrize('operator', OPERATOR_METHODS.keys())
def test_get_operator_method(operator):
    series = pd.Series([1, 2, 3, 4, 5])
    pd_operator = get_operator_method(operator)
    result = getattr(series, pd_operator)(series)
    assert ptypes.is_bool_dtype(result)


@pytest.mark.usefixtures("data", "rules")
class TestSimpleThresholder(TestCase):
    def test__convert_string_rule_to_dict(self):
        thresholder = SimpleThresholder(self.rules, "or")
        results = thresholder._convert_string_rule_to_dict(self.rules[0])
        expected_results = {"feature_name": "x1", "operator": "gt", "threshold": 0}
        assert results == expected_results

    def test_all_feature_names_property(self):
        thresholder = SimpleThresholder(self.rules, "or")
        assert thresholder.all_feature_names == ["x1", "x2"]

    def test_fit(self):
        thresholder = SimpleThresholder(self.rules, "or")
        assert thresholder.feature_importances_ is None
        thresholder.fit(x=self.data["X_train"], y=self.data["y_train"])
        np.testing.assert_array_equal(
            thresholder.feature_importances_, np.array([1, 1, 0, 0, 0, 0])
        )

    def test_rule_with_unavailable_feature_raises_error(self):
        rules = self.rules
        rules.append("x5 == 3")
        thresholder = SimpleThresholder(rules, "or")
        with self.assertRaises(BaselineFeatureNotInMatrix):
            thresholder.fit(x=self.data["X_train"], y=self.data["y_train"])

    def test_predict_proba(self):
        for logical_operator in ["and", "or"]:
            thresholder = SimpleThresholder(self.rules, logical_operator)
            thresholder.fit(x=self.data["X_train"], y=self.data["y_train"])
            results = thresholder.predict_proba(self.data["X_test"])
            if logical_operator == "and":
                expected_results = np.array(
                    [[0, 1, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0]]
                ).transpose()
            elif logical_operator == "or":
                expected_results = np.array(
                    [[1, 1, 1, 1, 1, 1, 0, 1], [1, 1, 1, 1, 1, 1, 0, 1]]
                ).transpose()
            np.testing.assert_array_equal(results, expected_results)


@pytest.mark.usefixtures("data", "features", "weights")
class TestLinearRanker(TestCase):
    def test_fit(self):
        ranker = LinearRanker(
            self.features,
            self.weights
        )
        assert ranker.feature_importances_ is None
        ranker.fit(x=self.data["X_train"], y=self.data["y_train"])
        np.testing.assert_almost_equal(
            ranker.feature_importances_, 
            np.array([0.19354839, 0.16129032, 
                      0.12903226, 0.22580645, 
                      0.19354839, 0.09677419]),
            decimal=4
        )

    def test_predict_proba(self):
        ranker = LinearRanker(self.features, 
                              self.weights)
        ranker.fit(x=self.data["X_train"], y=self.data["y_train"])
        results = ranker.predict_proba(self.data["X_test"])
        
        assert results.shape == (8, 2)
        
        expected_results = np.array([[0.75411428, 0.24588572],
                                     [0.69980399, 0.30019601],
                                     [0.79451361, 0.20548639],
                                     [0.27090783, 0.72909217],
                                     [0.39821089, 0.60178911],
                                     [0.67697988, 0.32302012],
                                     [0.51544567, 0.48455433],
                                     [0.63648246, 0.36351754]])

        np.testing.assert_almost_equal(results, 
                                       expected_results, 
                                       decimal=4)
        



"""Break ties in list predictions

Revision ID: ce5b50ffa8e2
Revises: 264786a9fe85
Create Date: 2021-01-08 21:59:13.403934

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'ce5b50ffa8e2'
down_revision = '264786a9fe85'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('list_predictions', sa.Column('rank_abs_with_ties', sa.Integer(), nullable=True), schema='production')
    op.add_column('list_predictions', sa.Column('rank_pct_with_ties', sa.Float(), nullable=True), schema='production')
    op.alter_column('list_predictions', 'rank_abs', new_column_name='rank_abs_no_ties', schema='production')
    op.alter_column('list_predictions', 'rank_pct', new_column_name='rank_pct_no_ties', schema='production')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('list_predictions', 'rank_abs_no_ties', new_column_name='rank_abs', schema='production')
    op.alter_column('list_predictions', 'rank_pct_no_ties', new_column_name='rank_pct', schema='production')
    op.drop_column('list_predictions', 'rank_pct_with_ties', schema='production')
    op.drop_column('list_predictions', 'rank_abs_with_ties', schema='production')
    # ### end Alembic commands ###



"""add hashes to runs

Revision ID: 3ce027594a5c
Revises: 5dd2ba8222b1
Create Date: 2022-03-25 12:58:38.370271

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3ce027594a5c'
down_revision = '5dd2ba8222b1'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('triage_runs', sa.Column('cohort_table_name', sa.String(), nullable=True), schema='triage_metadata')
    op.add_column('triage_runs', sa.Column('labels_table_name', sa.String(), nullable=True), schema='triage_metadata')
    op.add_column('triage_runs', sa.Column('bias_hash', sa.String(), nullable=True), schema='triage_metadata')


def downgrade():
    op.drop_column('triage_runs', 'bias_hash', schema='triage_metadata')
    op.drop_column('triage_runs', 'labels_table_name', schema='triage_metadata')
    op.drop_column('triage_runs', 'cohort_table_name', schema='triage_metadata')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re
from pathlib import Path
from setuptools import find_packages, setup


ROOT_PATH = Path(__file__).parent

LICENSE_PATH = ROOT_PATH / "LICENSE"

README_PATH = ROOT_PATH / "README.md"

REQUIREMENTS_PATH = ROOT_PATH / "requirement" / "main.txt"

REQUIREMENTS_TEST_PATH = ROOT_PATH / "requirement" / "test.txt"

REQUIREMENTS_RQ_PATH = ROOT_PATH / "requirement" / "extras-rq.txt"


def stream_requirements(fd):
    """For a given requirements file descriptor, generate lines of
    distribution requirements, ignoring comments and chained requirement
    files.

    """
    for line in fd:
        cleaned = re.sub(r"#.*$", "", line).strip()
        if cleaned and not cleaned.startswith("-r"):
            yield cleaned


with REQUIREMENTS_PATH.open() as requirements_file:
    REQUIREMENTS = list(stream_requirements(requirements_file))


with REQUIREMENTS_TEST_PATH.open() as test_requirements_file:
    REQUIREMENTS_TEST = REQUIREMENTS[:]
    REQUIREMENTS_TEST.extend(stream_requirements(test_requirements_file))

with REQUIREMENTS_RQ_PATH.open() as rq_requirements_file:
    RQ_REQUIREMENTS = list(stream_requirements(rq_requirements_file))


setup(
    name='triage',
    version='5.5.4',
    description="Risk modeling and prediction",
    long_description=README_PATH.read_text(),
    long_description_content_type="text/markdown",
    author="Center for Data Science and Public Policy",
    author_email="datascifellows@gmail.com",
    url="https://dssg.github.io/triage/",
    project_urls={
        "Documentation": "https://dssg.github.io/triage/",
        "Source Code": "https://github.com/dssg/triage",
        "Tutorial": "https://dssg.github.io/triage/dirtyduck/",
    },
    packages=find_packages("src", exclude=["tests", "tests.*"]),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=REQUIREMENTS,
    entry_points={
        "console_scripts": ["triage = triage.cli:execute"],
    },
    extras_require={"rq": RQ_REQUIREMENTS},
    license="MIT License",
    zip_safe=False,
    keywords="triage",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires=">=3.8",
    test_suite="tests",
    tests_require=REQUIREMENTS_TEST,
)



import verboselogs

from triage.database_reflection import table_has_data, table_row_count, table_exists, table_has_duplicates


logger = verboselogs.VerboseLogger(__name__)
DEFAULT_ACTIVE_STATE = "active"


class EntityDateTableGenerator:
    """Create a table containing state membership on different dates

    The structure of the output table is:
        entity_id
        date
        active (boolean): Whether or not the entity is considered 'active'
            (i.e., in the cohort or subset) on that date

    Args:
        db_engine (sqlalchemy.engine)
        experiment_hash (string) unique identifier for the experiment
        query (string) SQL query string to select entities for a given as_of_date
            The as_of_date should be parameterized with brackets: {as_of_date}
        replace (boolean) Whether or not to overwrite old rows.
            If false, each as-of-date will query to see if there are existing rows
                and not run the query if so.
            If true, the existing table will be dropped and recreated.
    """
    def __init__(self, query, db_engine, entity_date_table_name, labels_table_name=None, replace=True):
        self.db_engine = db_engine
        self.query = query
        self.entity_date_table_name = entity_date_table_name
        self.labels_table_name = labels_table_name
        self.replace = replace

    def generate_entity_date_table(self, as_of_dates):
        """Convert the object's input table
        into a states table for the given as_of_dates

        Args:
            as_of_dates (list of datetime.dates) Dates to include in the
                state table
        """
        logger.spam(f"Generating entity_date table {self.entity_date_table_name}")
        if self.query:
            logger.spam(f"Query is present, so running query on as_of_dates: {as_of_dates}")
            self._create_and_populate_entity_date_table_from_query(as_of_dates)
        elif self.labels_table_name:
            self._create_and_populate_entity_date_table_from_labels()
        else:
            raise ValueError("Neither query not labels table name is available, cannot compute cohort")
        logger.spam(f"Table {self.entity_date_table_name} created and populated")

        if not table_has_data(self.entity_date_table_name, self.db_engine):
            raise ValueError(self._empty_table_message(as_of_dates))

        if table_has_duplicates(
            self.entity_date_table_name,
            ['entity_id', 'as_of_date'],
            self.db_engine
            ):
            raise ValueError(f"Duplicates found in {self.entity_date_table_name}!")

        logger.debug(f"Entity-date table generated at {self.entity_date_table_name}")
        logger.spam(f"Generating stats on {self.entity_date_table_name}")
        logger.spam(f"Row count of {self.entity_date_table_name}: {table_row_count(self.entity_date_table_name, self.db_engine)}")


    def _maybe_create_entity_date_table(self):
        if self.replace or not table_exists(self.entity_date_table_name, self.db_engine):
            logger.spam(f"Creating entity_date table {self.entity_date_table_name}")
            self.db_engine.execute(f"drop table if exists {self.entity_date_table_name}")
            self.db_engine.execute(
                f"""create table {self.entity_date_table_name} (
                    entity_id integer,
                    as_of_date timestamp,
                    {DEFAULT_ACTIVE_STATE} boolean
                )
                """
            )

            logger.spam(f"Creating indices on entity_id and as_of_date for entity_date table {self.entity_date_table_name}")
            self.db_engine.execute(
                f"create index on {self.entity_date_table_name} (entity_id, as_of_date)"
            )
        else:
            logger.notice(
                f"Not dropping and recreating entity_date {self.entity_date_table_name} table because "
                f"replace flag was set to False and table was found to exist"
            )

    def _create_and_populate_entity_date_table_from_query(self, as_of_dates):
        """Create an entity_date table by sequentially running a
            given date-parameterized query for all known dates.

        Args:
        as_of_dates (list of datetime.date): Dates to calculate entity states as of
        """
        
        self._maybe_create_entity_date_table()
        logger.spam(f"Inserting rows into entity_date table {self.entity_date_table_name}")
        for as_of_date in as_of_dates:
            formatted_date = f"{as_of_date.isoformat()}"
            logger.spam(f"Looking for existing entity_date rows for as of date {as_of_date}")
            any_existing = list(self.db_engine.execute(
                f"""select 1 from {self.entity_date_table_name}
                where as_of_date = '{formatted_date}'
                limit 1
                """
            ))
            if len(any_existing) == 1:
                logger.notice(f"Since >0 entity_date rows found for date {as_of_date}, skipping")
                continue
            dated_query = self.query.format(as_of_date=formatted_date)
            full_query = f"""insert into {self.entity_date_table_name}
                select q.entity_id, '{formatted_date}'::timestamp, true
                from ({dated_query}) q
                group by 1, 2, 3
            """
            logger.spam(f"Running entity_date query for date: {as_of_date}, {full_query}")
            self.db_engine.execute(full_query)

    def _create_and_populate_entity_date_table_from_labels(self):
        """Create an entity_date table by storing all distinct entity-id/as-of-date pairs
        from the labels table
        """
        self._maybe_create_entity_date_table()
        logger.spam(f"Populating entity_date table {self.entity_date_table_name} from labels table {self.labels_table_name}")
        if not table_exists(self.labels_table_name, self.db_engine):
            logger.warning("Labels table does not exist, cannot populate entity-dates")
            return

        # If any rows exist in the entity_date table, don't insert any for dates
        # already in the table. This replicates the logic used above by
        # _create_and_populate_entity_date_table_from_query
        logger.spam(f"Looking for existing entity_date rows for label as of dates")
        existing_dates = list(self.db_engine.execute(
            f"""
            with label_dates as (
                select distinct as_of_date::DATE AS as_of_date FROM {self.labels_table_name}
            )
            , cohort_dates as (
                select distinct as_of_date::DATE AS as_of_date FROM {self.entity_date_table_name}
            )
            select distinct l.as_of_date
            from label_dates l
            join cohort_dates c using(as_of_date)
            """
        ))
        if len(existing_dates) > 0:
            existing_dates = ', '.join([rec[0].isoformat() for rec in existing_dates])
            logger.notice(f'Existing entity_dates records found for the following dates, '
                f'so new records will not be inserted for these dates {existing_dates}')

        insert_query = f"""
            insert into {self.entity_date_table_name}
            select distinct entity_id, as_of_date, true
            from (
                select distinct l.entity_id, l.as_of_date
                from {self.labels_table_name} as l
                left join (select distinct as_of_date from {self.entity_date_table_name}) as c
                    on l.as_of_date::DATE = c.as_of_date::DATE
                where c.as_of_date IS NULL
            ) as sub
        """
        logger.spam(f"Running entity_date query from labels table: {insert_query}")
        self.db_engine.execute(insert_query)

    def _empty_table_message(self, as_of_dates):
        return """Query does not return any rows for the given as_of_dates:
            {as_of_dates}
            '{query}'""".format(
            query=self.query or "labels table",
            as_of_dates=", ".join(
                str(as_of_date)
                for as_of_date in (
                    as_of_dates if len(as_of_dates) <= 5 else as_of_dates[:5] + ["…"]
                )
            ),
        )

    def clean_up(self):
        self.db_engine.execute(f"drop table if exists {self.entity_date_table_name}")


class CohortTableGeneratorNoOp(EntityDateTableGenerator):
    def __init__(self):
        pass

    def generate_entity_date_table(self, as_of_dates):
        logger.warning(
            "No cohort configuration is available, so no cohort will be created"
        )
        return

    def clean_up(self):
        logger.warning("No cohort configuration is available, so no cohort will be tear down")
        return

    @property
    def entity_date_table_name(self):
        return None


class SubsetEntityDateTableGenerator(EntityDateTableGenerator):
    def __init__(self, query, db_engine, entity_date_table_name, labels_table_name=None, replace=True, cohort_table=None):
        super().__init__(query, db_engine, entity_date_table_name, labels_table_name, replace)
        print('Initializing the new child class Subset entity date generator')
        self.cohort_table = cohort_table
        
    def create_and_populate_entity_date_table_from_query(self, as_of_dates):
        """Create an entity_date table by sequentially running a
            given date-parameterized query for all known dates.

        Args:
            as_of_dates (list of datetime.date): Dates to calculate entity states as of
        """
        
        self._maybe_create_entity_date_table()
        logger.spam(f"Inserting rows into entity_date table {self.entity_date_table_name}")
        
        for as_of_date in as_of_dates:
            formatted_date = f"{as_of_date.isoformat()}"
            logger.spam(f"Looking for existing entity_date rows for as of date {as_of_date}")
            any_existing = list(self.db_engine.execute(
                f"""select 1 from {self.entity_date_table_name}
                where as_of_date = '{formatted_date}'
                limit 1
                """
            ))
            if len(any_existing) == 1:
                logger.notice(f"Since >0 entity_date rows found for date {as_of_date}, skipping")
                continue
            dated_query = self.query.format(as_of_date=formatted_date)
            full_query = f"""insert into {self.entity_date_table_name}
                select q.entity_id, '{formatted_date}'::timestamp, true
                from (
                    with subset as ({dated_query})
                    select 
                        c. entity_id
                    from subset s inner join {self.cohort_table} c
                    on s.entity_id = c.entity_id
                    and c.as_of_date = '{formatted_date}'::date
                ) q 
                group by 1, 2, 3
            """
            logger.spam(f"Running entity_date query for date: {as_of_date}, {full_query}")
            self.db_engine.execute(full_query)
            
            
    def generate_entity_date_table(self, as_of_dates):
        if self.query:
            logger.spam(f"Query is present, so running query on as_of_dates: {as_of_dates}")
            self.create_and_populate_entity_date_table_from_query(as_of_dates)
        else:
            logger.warning('Query not working, subset table not created!')
    
        if table_has_duplicates(
            self.entity_date_table_name,
            ['entity_id', 'as_of_date'],
            self.db_engine
            ):
            raise ValueError(f"Duplicates found in {self.entity_date_table_name}!")

        logger.debug(f"Entity-date table generated at {self.entity_date_table_name}")
        logger.spam(f"Generating stats on {self.entity_date_table_name}")
        logger.spam(f"Row count of {self.entity_date_table_name}: {table_row_count(self.entity_date_table_name, self.db_engine)}")


import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

import datetime
import shutil
import sys
import random
from contextlib import contextmanager
import functools
import operator
import tempfile
import subprocess
import gzip

import sqlalchemy

import pandas as pd
import numpy as np
from sqlalchemy.orm import sessionmaker

from triage.component.results_schema import Model
from triage.util.structs import FeatureNameList


def str_in_sql(values):
    return ",".join(map(lambda x: "'{}'".format(x), values))


def feature_list(feature_dictionary):
    """Convert a feature dictionary to a sorted list

    Args: feature_dictionary (dict)

    Returns: sorted list of feature names
    """
    if not feature_dictionary:
        return FeatureNameList()
    return FeatureNameList(sorted(
        functools.reduce(
            operator.concat,
            (feature_dictionary[key] for key in feature_dictionary.keys()),
        )
    ))


def convert_string_column_to_date(column):
    return [datetime.datetime.strptime(date, "%Y-%m-%d").date() for date in column]


def create_features_table(table_number, table, engine):
    engine.execute(
        """
            create table features.features{} (
                entity_id int, as_of_date date, f{} int, f{} int
            )
        """.format(
            table_number, (table_number * 2) + 1, (table_number * 2) + 2
        )
    )
    for row in table:
        engine.execute(
            """
                insert into features.features{} values (%s, %s, %s, %s)
            """.format(
                table_number
            ),
            row,
        )


def create_entity_date_df(
    labels,
    states,
    as_of_dates,
    state_one,
    state_two,
    label_name,
    label_type,
    label_timespan,
):
    """ This function makes a pandas DataFrame that mimics the entity-date table
    for testing against.
    """
    0, "2016-02-01", "1 month", "booking", "binary", 0
    labels_table = pd.DataFrame(
        labels,
        columns=[
            "entity_id",
            "as_of_date",
            "label_timespan",
            "label_name",
            "label_type",
            "label",
        ],
    )
    states_table = pd.DataFrame(
        states, columns=["entity_id", "as_of_date", "state_one", "state_two"]
    ).set_index(["entity_id", "as_of_date"])
    as_of_dates = [date.date() for date in as_of_dates]
    labels_table = labels_table[labels_table["label_name"] == label_name]
    labels_table = labels_table[labels_table["label_type"] == label_type]
    labels_table = labels_table[labels_table["label_timespan"] == label_timespan]
    labels_table = labels_table.join(other=states_table, on=("entity_id", "as_of_date"))
    labels_table = labels_table[labels_table["state_one"] & labels_table["state_two"]]
    ids_dates = labels_table[["entity_id", "as_of_date"]]
    ids_dates = ids_dates.sort_values(["entity_id", "as_of_date"])
    ids_dates["as_of_date"] = [
        datetime.datetime.strptime(date, "%Y-%m-%d").date()
        for date in ids_dates["as_of_date"]
    ]
    ids_dates = ids_dates[ids_dates["as_of_date"].isin(as_of_dates)]
    logger.spam(ids_dates)

    return ids_dates.reset_index(drop=True)


def change_datetimes_on_metadata(metadata):
    for element in metadata.keys(): 
        if (element.endswith("_time")) or (element.endswith("_times")):
            if isinstance(metadata[element], list):
                metadata[element] = [str(ele) for ele in metadata[element]]
            else: 
                metadata[element] = str(metadata[element])
   
    return metadata


def NamedTempFile():
    if sys.version_info >= (3, 0, 0):
        return tempfile.NamedTemporaryFile(mode="w+", newline="")
    else:
        return tempfile.NamedTemporaryFile()


@contextmanager
def TemporaryDirectory():
    name = tempfile.mkdtemp()
    try:
        yield name
    finally:
        shutil.rmtree(name)


def fake_labels(length):
    return np.array([random.choice([True, False]) for i in range(0, length)])


class MockTrainedModel:
    def predict_proba(self, dataset):
        return np.random.rand(len(dataset), len(dataset))


def fake_trained_model(project_path, model_storage_engine, db_engine):
    """Creates and stores a trivial trained model

    Args:
        project_path (string) a desired fs/s3 project path
        model_storage_engine (triage.storage.ModelStorageEngine)
        db_engine (sqlalchemy.engine)

    Returns:
        (int) model id for database retrieval
    """
    trained_model = MockTrainedModel()
    model_storage_engine.write(trained_model, "abcd")
    session = sessionmaker(db_engine)()
    db_model = Model(model_hash="abcd")
    session.add(db_model)
    session.commit()
    return trained_model, db_model.model_id


def assert_index(engine, table, column):
    """Assert that a table has an index on a given column

    Does not care which position the column is in the index
    Modified from https://www.gab.lc/articles/index_on_id_with_postgresql

    Args:
        engine (sqlalchemy.engine) a database engine
        table (string) the name of a table
        column (string) the name of a column
    """
    query = """
        SELECT 1
        FROM pg_class t
             JOIN pg_index ix ON t.oid = ix.indrelid
             JOIN pg_class i ON i.oid = ix.indexrelid
             JOIN pg_attribute a ON a.attrelid = t.oid
        WHERE
             a.attnum = ANY(ix.indkey) AND
             t.relkind = 'r' AND
             t.relname = '{table_name}' AND
             a.attname = '{column_name}'
    """.format(
        table_name=table, column_name=column
    )
    num_results = len([row for row in engine.execute(query)])
    assert num_results >= 1


def create_dense_state_table(db_engine, table_name, data):
    db_engine.execute(
        """create table {} (
        entity_id int,
        state text,
        start_time timestamp,
        end_time timestamp
    )""".format(
            table_name
        )
    )

    for row in data:
        db_engine.execute(
            "insert into {} values (%s, %s, %s, %s)".format(table_name), row
        )


def create_binary_outcome_events(db_engine, table_name, events_data):
    db_engine.execute(
        "create table events (entity_id int, outcome_date date, outcome bool)"
    )
    for event in events_data:
        db_engine.execute(
            "insert into {} values (%s, %s, %s::bool)".format(table_name), event
        )


def retry_if_db_error(exception):
    return isinstance(exception, sqlalchemy.exc.OperationalError)


def _num_elements(x):
    """Extract the number of rows from the subprocess output"""
    return int(str(x.stdout, encoding="utf-8").split(" ")[0])


def check_rows_in_files(filenames, matrix_uuid):
    """Checks if the number of rows among all the CSV files for features and 
    and label for a matrix uuid are the same. 

    Args:
        filenames (List): List of CSV files to check the number of rows
        path_ (string): Path to get the temporal csv files
    """
    outputs = []
    for element in filenames:
        logging.debug(f"filename: {element}")
        just_filename = element.split("/")[-1]
        if (element.endswith(".csv")) and (just_filename.startswith(matrix_uuid)):
            cmd_line = "wc -l " + element
            outputs.append(subprocess.run(cmd_line, shell=True, capture_output=True))

    # get the number of rows from the subprocess
    rows = [_num_elements(output) for output in outputs]
    rows_set = set(rows)
    logging.debug(f"number of rows in files {rows_set}")

    if len(rows_set) == 1: 
        return True
    else:
        return False

def check_entity_ids_in_files(filenames, matrix_uuid):
    """Verifies if all the files in features and label have the same exact entity ids and knowledge dates"""
    # get first 2 columns on each file (entity_id, knowledge_date)
    for element in filenames: 
        logging.debug(f"getting entity id and knowledge date from features {element}")
        just_filename = element.split("/")[-1]
        prefix = element.split(".")[0]
        if (element.endswith(".csv")) and (just_filename.startswith(matrix_uuid)):
            cmd_line = f"cut -d ',' -f 1,2 {element} | sort -k 1,2 > {prefix}_sorted.csv"
            subprocess.run(cmd_line, shell=True)
    
    base_file = filenames[0]
    comparisons = []
    for i in range(1, len(filenames)):
        if (filenames[i].endswith(".csv")) and (filenames[i].startswith(matrix_uuid)):
            cmd_line = f"diff {base_file} {filenames[i]}"
            comparisons.append(subprocess.run(cmd_line, shell=True, capture_output=True))
    
    if len(comparisons) == 0:
        return True
    else:
        return False


def remove_entity_id_and_knowledge_dates(filenames, matrix_uuid):
    """drop entity id and knowledge date from all features and label files but one""" 
    correct_filenames = []

    for i in range(len(filenames)):
        just_filename = filenames[i].split("/")[-1]
        prefix = filenames[i].split(".")[0]
        if not (just_filename.endswith("_sorted.csv")) and (just_filename.startswith(matrix_uuid)):
            if prefix.endswith("_0"): 
                # only the first file will have entity_id and knowledge data but needs to also be sorted
                cmd_line = f"sort -k 1,2 {filenames[i]} > {prefix}_fixed.csv"
            else:
                cmd_line = f"sort -k 1,2 {filenames[i]} | cut -d ',' -f 3- > {prefix}_fixed.csv"
            subprocess.run(cmd_line, shell=True)
            # all files now the header in the last row (after being sorted)
            # from https://www.unix.com/shell-programming-and-scripting/128416-use-sed-move-last-line-top.html
            # move last line to first line
            cmd_line = f"sed -i '1h;1d;$!H;$!d;G' {prefix}_fixed.csv"
            subprocess.run(cmd_line, shell=True)
            correct_filenames.append(f"{prefix}_fixed.csv")

    return correct_filenames

    
def generate_list_of_files_to_remove(filenames, matrix_uuid):
    """Generate the list of all files that need to be removed"""
    # adding _sorted
    rm_files = []

    for element in filenames:
        rm_files.append(element)
        if (element.split("/")[-1].startswith(matrix_uuid)):
            prefix = element.split(".")[0]
            # adding sorted files 
            rm_files.append(prefix + "_sorted.csv")
            # adding fixed files
            rm_files.append(prefix + "_fixed.csv")

    logging.debug(f"Files to be removed {rm_files}")
    return rm_files


def generate_gzip(path, matrix_uuid):
    """
    Generates a gzip from the csv file with all the features (doesn't include the label)

    Args:
        path (string): _description_
        matrix_uuid (string): _description_
    """
    filename_ = f"{path}/{matrix_uuid}.csv"
    filepath_ = "/".join(filename_.split("/")[:-1])
    logger.debug(f"About to generate gzip for {matrix_uuid}.csv in {filepath_}")
    try:
        with open(filename_, 'rb') as f_in:
            with gzip.open(f"{filename_}.gz", 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
                logger.debug(f"gzip file {filename_}.gz generated")
    except FileNotFoundError as e:
        logger.error(f"File {filename_} not found: {e}")
    except Exception as e:
        logger.error(f"An error occurred while generating gzip: {e}")


def remove_unnecessary_files(filenames, path_, matrix_uuid, storage_is_s3):
    """
    Removes the csvs generated for each feature, the label csv file,
    and the csv with all the features and label stitched togheter. 
    The csv with all merged is being deleted while generating the gzip.

    Args:
        filenames (list): list of filenames to remove from disk
        path_ (string): Path 
        matrix_uuid (string): ID of the matrix
        storage_is_s3 (bool): Whether the project store is a S3 bucket
    """
    # deleting features and label csvs
    for filename_ in filenames:
        cmd_line = ['rm', filename_] 
        logger.debug(f"removing files with command {cmd_line}")
        try:
            subprocess.run(cmd_line, check=True)
        except subprocess.CalledProcessError as e:
            logger.error(f"Error removing file {filename_}: {e}")

    # deleting the merged csv
    cmd_line = ['rm', f'{path_}/{matrix_uuid}.csv']
    logger.debug(f"removing stitched csv with command {cmd_line}")
    try:
        subprocess.run(cmd_line, check=True)
    except subprocess.CalledProcessError as e:
        logger.error(f"Error removing file {path_}/{matrix_uuid}.csv: {e}")
    
    # deleting the compressed CSV when the project path is S3
    if storage_is_s3:
        filename_ = f"{path_}/{matrix_uuid}.csv.gz"
        cmd_line = ['rm', filename_]
        logger.debug(f"About to remove gzip file with command {cmd_line}")
        try:
            subprocess.run(cmd_line, check=True)
            logger.debug(f"gzip file {filename_} removed")
        except subprocess.CalledProcessError as e:
            logger.error(f"Error removing file {filename_}: {e}")
            




import pandas as pd

from triage.component.postmodeling.crosstabs import run_crosstabs, run_crosstabs_from_matrix
from triage.database_reflection import table_has_data


def test_run_crosstabs(finished_experiment, crosstabs_config):
    run_crosstabs(finished_experiment.db_engine, crosstabs_config)
    expected_table_name = (
        crosstabs_config.output["schema"] + "." + crosstabs_config.output["table"]
    )
    assert table_has_data(expected_table_name, finished_experiment.db_engine)

    
def test_run_crosstabs_from_matrix(finished_experiment):
    """assert that crosstabs table is populated for all threshold types"""
    
    thresholds = {
        'rank_abs_no_ties': 50,
        'rank_abs_with_ties': 50, 
        'rank_pct_no_ties': 0.1,
        'rank_pct_with_ties': 0.1
    }
    table_schema='test_results'
    table_name = 'crosstabs'
    
    errors = list()
    for threshold_type, threshold in thresholds.items():    
        df = run_crosstabs_from_matrix(
            db_engine=finished_experiment.db_engine,
            project_path=finished_experiment.project_path,
            model_id=1,
            threshold_type=threshold_type,
            threshold=threshold,
            push_to_db=True,
            table_schema=table_schema,
            table_name=table_name,
            replace=True
        )
        
        q = f'''
            select 
            1
            from {table_schema}.{table_name}
            where model_id = 1
            and threshold_type = '{threshold_type}'
            and threshold = {threshold}
        '''
        
        df = pd.read_sql(q,finished_experiment.db_engine)
                
        if df.empty:
            errors.append(threshold_type)
 
    assert not errors, f"errors occured for: {errors}"    



import pytest

from triage.component.postmodeling.add_predictions import add_predictions
from triage.database_reflection import table_has_data


MODEL_IDS_QUERY = """
    SELECT
        model_id
    FROM triage_metadata.models
    WHERE model_group_id={model_group_id}
"""

MODELS_IN_PREDICTIONS_QUERY = """
    SELECT 
        distinct model_id
    FROM test_results.predictions
"""


def test_populate_predictions_table(finished_experiment_without_predictions):
    """assert that generate_predictions populate the predictions table"""

    db_engine = finished_experiment_without_predictions.db_engine
    model_groups = [1]
    project_path = finished_experiment_without_predictions.project_storage.project_path

    add_predictions(
        db_engine=db_engine,
        model_groups=model_groups,
        project_path=project_path
    )

    assert table_has_data('test_results.predictions', db_engine)


def test_add_predictions_all_models(finished_experiment_without_predictions):
    """assert that generate_predictions write predictions of all models in the model group"""

    db_engine = finished_experiment_without_predictions.db_engine
    model_groups = [1]
    project_path = finished_experiment_without_predictions.project_storage.project_path

    add_predictions(
        db_engine=db_engine,
        model_groups=model_groups,
        project_path=project_path
    )

    # Model ids belonging to the model group 1
    model_ids = db_engine.execute(MODEL_IDS_QUERY.format(model_group_id=1)).fetchall()
    model_ids = {x[0] for x in model_ids}

    # model ids present in the predictions table
    model_ids_predictions = db_engine.execute(MODELS_IN_PREDICTIONS_QUERY).fetchall()
    model_ids_predictions = {x[0] for x in model_ids_predictions}
        
    assert model_ids == model_ids_predictions


def test_models_invalid_model_group(finished_experiment_without_predictions):
    """Test whether the module produces the error for invalid model groups"""

    db_engine = finished_experiment_without_predictions.db_engine
    model_groups = [325]
    project_path = finished_experiment_without_predictions.project_storage.project_path

    with pytest.raises(ValueError):
        add_predictions(
            db_engine=db_engine,
            model_groups=model_groups,
            project_path=project_path
        )


def test_models_not_found_invalid_time_range(finished_experiment_without_predictions):
    """Test whether the module errors when an invalid time range is provided"""
    db_engine = finished_experiment_without_predictions.db_engine
    model_groups = [1]
    project_path = finished_experiment_without_predictions.project_storage.project_path

    q = """
        SELECT 
            to_char(min(train_end_time) - '1month'::interval, 'YYYY-MM-DD') as lower_lim,
            to_char(max(train_end_time) + '1month'::interval, 'YYYY-MM-DD') as upper_lim
        FROM triage_metadata.models
        where model_group_id = 1

    """

    end_times = db_engine.execute(q).fetchall()[0]
    lower_lim = end_times[0]
    upper_lim = end_times[1]

    with pytest.raises(ValueError):
        add_predictions(
            db_engine=db_engine,
            model_groups=model_groups,
            project_path=project_path,
            train_end_times_range={
                'range_start_date': upper_lim,
            }
        )

    with pytest.raises(ValueError):
        add_predictions(
            db_engine=db_engine,
            model_groups=model_groups,
            project_path=project_path,
            train_end_times_range={
                'range_end_date': lower_lim
            }
        )

    with pytest.raises(ValueError):
        add_predictions(
            db_engine=db_engine,
            model_groups=model_groups,
            project_path=project_path,
            train_end_times_range={
                'range_start_date': upper_lim,
                'range_end_date': upper_lim
            }
        )


def test_invalid_experiment_hash(finished_experiment_without_predictions):
    """Test whether an invalid experiment hash throws a ValueError"""

    db_engine = finished_experiment_without_predictions.db_engine
    model_groups = [1]
    project_path = finished_experiment_without_predictions.project_storage.project_path

    experiment_hash = finished_experiment_without_predictions.experiment_hash
    experiment_hash = experiment_hash + 'x'

    with pytest.raises(ValueError):
        add_predictions(
            db_engine=db_engine,
            model_groups=model_groups,
            project_path=project_path,
            experiment_hashes=[experiment_hash]
        )



import io
import subprocess

import verboselogs, logging
logger = verboselogs.VerboseLogger(__name__)

import pandas as pd
import numpy as np
import polars as pl
import time

from sqlalchemy.orm import sessionmaker

from triage.component.results_schema import Matrix
from triage.database_reflection import table_has_data, table_row_count
from triage.tracking import built_matrix, skipped_matrix, errored_matrix
from triage.util.pandas import downcast_matrix
from triage.component.architect.utils import (
    change_datetimes_on_metadata, 
    check_rows_in_files,
    check_entity_ids_in_files,
    remove_entity_id_and_knowledge_dates,
    generate_list_of_files_to_remove,
    generate_gzip,
    remove_unnecessary_files,
)
from triage.component.catwalk.storage import S3Store

class BuilderBase:
    def __init__(
        self,
        db_config,
        matrix_storage_engine,
        engine,
        experiment_hash,
        replace=True,
        include_missing_labels_in_train_as=None,
        run_id=None,
    ):
        self.db_config = db_config
        self.matrix_storage_engine = matrix_storage_engine
        self.db_engine = engine
        self.experiment_hash = experiment_hash
        self.replace = replace
        self.include_missing_labels_in_train_as = include_missing_labels_in_train_as
        self.run_id = run_id
        self.includes_labels = 'labels_table_name' in self.db_config

    @property
    def sessionmaker(self):
        return sessionmaker(bind=self.db_engine)

    def validate(self):
        for expected_db_config_val in [
            "features_schema_name",
            "cohort_table_name",
            "labels_schema_name",
            "labels_table_name",
        ]:
            if expected_db_config_val not in self.db_config:
                raise ValueError(
                    "{} needed in db_config".format(expected_db_config_val)
                )

    def build_all_matrices(self, build_tasks):
        logger.info(f"Building {len(build_tasks.keys())} matrices")

        for i, (matrix_uuid, task_arguments) in enumerate(build_tasks.items(), start=1):
            logger.info(
                f"Building matrix {matrix_uuid} [{i}/{len(build_tasks.keys())}]"
            )
            self.build_matrix(**task_arguments)
            logger.success(f"Matrix {matrix_uuid} built")

    def _outer_join_query(
        self,
        right_table_name,
        right_column_selections,
        entity_date_table_name,
        additional_conditions="",
        include_index=True,
        column_override=None,
    ):
        """ Given a (features or labels) table, a list of times, columns to
        select, and (optionally) a set of join conditions, perform an outer
        join to the entity date table.

        :param right_table_name: the name of the right (feature/label) table
        :param right_column_selections: formatted text for the columns to select
        :param entity_date_table_name: name of table containing all valid entity ids and dates
        :param additional_conditions: formatted text for additional join
                                      conditions
        :type right_table_name: str
        :type right_column_selections: str
        :type entity_date_table_name: str
        :type additional_conditions: str

        :return: postgresql query for the outer join to the entity-dates table
        :rtype: str
        """

        # put everything into the query
        if include_index:
            query = f"""
                SELECT ed.entity_id,
                    ed.as_of_date{"".join(right_column_selections)}
                FROM {entity_date_table_name} ed
                LEFT OUTER JOIN {right_table_name} r
                ON ed.entity_id = r.entity_id AND
                ed.as_of_date = r.as_of_date
                {additional_conditions}
                ORDER BY ed.entity_id,
                        ed.as_of_date
            """
        else:
            query = f"""
                with r as (
                    SELECT ed.entity_id,
                           ed.as_of_date, {"".join(right_column_selections)[2:]}
                    FROM {entity_date_table_name} ed
                    LEFT OUTER JOIN {right_table_name} r
                    ON ed.entity_id = r.entity_id AND
                       ed.as_of_date = r.as_of_date
                       {additional_conditions}
                    ORDER BY ed.entity_id,
                             ed.as_of_date
                ) 
                select {"".join(right_column_selections)[2:] if not column_override else column_override} 
                from r
            """
        
        return query
    

    def _join_no_labels_case(
        self,
        right_column_selections,
        entity_date_table_name,
        include_index=True,
        column_override=None,
    ):
        """
            Generates the label query for predict forward, when there are not labels 
            associated with the matrix. We return 0's as the label. 

            :param right_column_selections: formatted text for the columns to select
            :param entity_date_table_name: name of table containing all valid entity ids and dates

            :type right_column_selections: str
            :type entity_date_table_name: str

            :return: postgresql query for the outer join to the entity-dates table
            :rtype: str
        """
        # put everything into the query
        if include_index:
            query = f"""
                SELECT ed.entity_id,
                    ed.as_of_date{"".join(right_column_selections)}
                FROM {entity_date_table_name} ed
                ORDER BY ed.entity_id,
                        ed.as_of_date
            """
        else:
            query = f"""
                with r as (
                    SELECT ed.entity_id,
                           ed.as_of_date, {"".join(right_column_selections)[2:]}
                    FROM {entity_date_table_name} ed
                    ORDER BY ed.entity_id,
                             ed.as_of_date
                ) 
                select {"".join(right_column_selections)[2:] if not column_override else column_override} 
                from r
            """
        
        return query
    

    def make_entity_date_table(
        self,
        as_of_times,
        label_name,
        label_type,
        state,
        matrix_type,
        matrix_uuid,
        label_timespan,
    ):
        """ Make a table containing the entity_ids and as_of_dates required for
        the current matrix.

        :param as_of_times: the times to be used for the current matrix
        :param label_name: name of the label to be used
        :param label_type: the type of label to be used
        :param state: the entity state to be used in the matrix
        :param matrix_type: the type (train/test) of matrix
        :param matrix_uuid: a unique id for the matrix
        :param label_timespan: the time timespan that labels in matrix will include
        :type as_of_times: list
        :type label_name: str
        :type label_type: str
        :type state: str
        :type matrix_type: str
        :type matrix_uuid: str
        :type label_timespan: str

        :return: table name
        :rtype: str
        """

        as_of_time_strings = [str(as_of_time) for as_of_time in as_of_times]
        if matrix_type == "test" or matrix_type == "production" or self.include_missing_labels_in_train_as is not None:
            indices_query = self._all_valid_entity_dates_query(
                as_of_time_strings=as_of_time_strings, state=state
            )
        elif matrix_type == "train":
            indices_query = self._all_labeled_entity_dates_query(
                as_of_time_strings=as_of_time_strings,
                state=state,
                label_name=label_name,
                label_type=label_type,
                label_timespan=label_timespan,
            )
        else:
            raise ValueError(f"Unknown matrix type passed: {matrix_type}")

        table_name = "_".join([matrix_uuid, "matrix_entity_date"])
        query = f"""
            DROP TABLE IF EXISTS {self.db_config["features_schema_name"]}."{table_name}";
            CREATE TABLE {self.db_config["features_schema_name"]}."{table_name}"
            AS ({indices_query})
        """
        logger.debug(
            f"Creating matrix-specific entity-date table for matrix {matrix_uuid} ",
        )
        logger.debug(f"with query {query}")
        self.db_engine.execute(query)

        return table_name

    def _all_labeled_entity_dates_query(
        self, as_of_time_strings, state, label_name, label_type, label_timespan
    ):
        query = f"""
            SELECT entity_id, as_of_date
            FROM {self.db_config["cohort_table_name"]}
            JOIN {self.db_config["labels_schema_name"]}.{self.db_config["labels_table_name"]} using (entity_id, as_of_date)
            WHERE {state}
            AND as_of_date IN (SELECT (UNNEST (ARRAY{as_of_time_strings}::timestamp[])))
            AND label_name = '{label_name}'
            AND label_type = '{label_type}'
            AND label_timespan = '{label_timespan}'
            AND label is not null
            ORDER BY entity_id, as_of_date
        """
        return query

    def _all_valid_entity_dates_query(self, state, as_of_time_strings):
        query = f"""
            SELECT entity_id, as_of_date
            FROM {self.db_config["cohort_table_name"]}
            WHERE {state}
            AND as_of_date IN (SELECT (UNNEST (ARRAY{as_of_time_strings}::timestamp[])))
            ORDER BY entity_id, as_of_date
        """
        if not table_has_data(
            self.db_config["cohort_table_name"], self.db_engine
        ):
            raise ValueError("Required cohort table does not exist")
        return query


class MatrixBuilder(BuilderBase):
    def build_matrix(
        self,
        as_of_times,
        label_name,
        label_type,
        feature_dictionary,
        matrix_metadata,
        matrix_uuid,
        matrix_type,
    ):
        """ Write a design matrix to disk with the specified paramters.

        :param as_of_times: datetimes to be included in the matrix
        :param label_name: name of the label to be used
        :param label_type: the type of label to be used
        :param feature_dictionary: a dictionary of feature tables and features
                                   to be included in the matrix
        :param matrix_metadata: a dictionary of metadata about the matrix
        :param matrix_uuid: a unique id for the matrix
        :param matrix_type: the type (train/test) of matrix
        :type as_of_times: list
        :type label_name: str
        :type label_type: str
        :type feature_dictionary: dict
        :type matrix_metadata: dict
        :type matrix_uuid: str
        :type matrix_type: str

        :return: none
        :rtype: none
        """
        logger.spam(f"popped matrix {matrix_uuid} build off the queue")
        if not table_has_data(
            self.db_config["cohort_table_name"], self.db_engine
        ):
            logger.warning("cohort table is not populated, cannot build matrix")
            if self.run_id:
                errored_matrix(self.run_id, self.db_engine)
            return

        if self.includes_labels:
            if not table_has_data(
                    f"{self.db_config['labels_schema_name']}.{self.db_config['labels_table_name']}",
                    self.db_engine,
            ):
                logger.warning("labels table is not populated, cannot build matrix")
                if self.run_id:
                    errored_matrix(self.run_id, self.db_engine)

        matrix_store = self.matrix_storage_engine.get_store(matrix_uuid)
        if not self.replace and matrix_store.exists:
            logger.notice(f"Skipping {matrix_uuid} because matrix already exists")
            if self.run_id:
                skipped_matrix(self.run_id, self.db_engine)
            return

        logger.debug(
            f'Storing matrix {matrix_metadata["matrix_id"]} in {matrix_store.matrix_base_store.path}'
        )
        # make the entity time table and query the labels and features tables
        logger.debug(f"Making entity date table for matrix {matrix_uuid}")
        try:
            entity_date_table_name = self.make_entity_date_table(
                as_of_times,
                label_name,
                label_type,
                matrix_metadata["state"],
                matrix_type,
                matrix_uuid,
                matrix_metadata.get("label_timespan", None),
            )
        except ValueError as e:
            logger.exception(
                "Not able to build entity-date table,  will not build matrix",
            )
            if self.run_id:
                errored_matrix(self.run_id, self.db_engine)
            return
        logger.spam(
            f"Extracting feature group data from database into file for matrix {matrix_uuid}"
        )
        
        feature_queries = self.feature_load_queries(feature_dictionary, entity_date_table_name)
        logger.debug(f"feature queries, number of queries: {len(feature_queries)}")
        
        # when predict forwarding we don't have labels
        logger.info(f"Includes labels?: {self.includes_labels}")
        
        label_query = self.label_load_query(
            label_name,
            label_type,
            entity_date_table_name,
            matrix_metadata["label_timespan"],
        )
        logger.debug(f"Label query while building the matrix: {label_query}")

        output, labels = self.stitch_csvs(feature_queries, label_query, matrix_store, matrix_uuid)
        logger.info(f"matrix stitched, pandas DF returned")
        matrix_store.metadata = matrix_metadata
        matrix_store.matrix_label_tuple = output, labels
        logger.info(f"Saving matrix metadata (yaml) for matrix {matrix_uuid}")
        matrix_store.save_matrix_metadata()

        # If completely archived, save its information to matrices table
        # At this point, existence of matrix already tested, so no need to delete from db
        logging.info(f"Getting all matrix metadata for matrix {matrix_uuid}")
        if matrix_type == "train":
            lookback = matrix_metadata["max_training_history"]
        else:
            lookback = matrix_metadata["test_duration"]

        row_count = table_row_count(
            '{schema}."{table}"'.format(
                schema=self.db_config["features_schema_name"],
                table=entity_date_table_name,
            ),
            self.db_engine
        )

        matrix = Matrix(
            matrix_id=matrix_metadata["matrix_id"],
            matrix_uuid=matrix_uuid,
            matrix_type=matrix_type,
            labeling_window=matrix_metadata["label_timespan"],
            num_observations=row_count[0], #row count is a tuple
            lookback_duration=lookback,
            feature_start_time=matrix_metadata["feature_start_time"],
            feature_dictionary=feature_dictionary,
            matrix_metadata=matrix_metadata,
            built_by_experiment=self.experiment_hash
        )
        logger.info(f"About to save all matrix metadata on DB for matrix {matrix_uuid}")
        # before saving the matrix metadata we need to cast datetimes to str 
        matrix_metadata = change_datetimes_on_metadata(matrix_metadata)
        session = self.sessionmaker()
        session.merge(matrix)
        session.commit()
        session.close()
        if self.run_id:
            built_matrix(self.run_id, self.db_engine)


    def label_load_query(
        self,
        label_name,
        label_type,
        entity_date_table_name,
        label_timespan,
    ):
        """ Query the labels table and write the data to disk in csv format.
        :param as_of_times: the times to be used for the current matrix
        :param label_name: name of the label to be used
        :param label_type: the type of label to be used
        :param entity_date_table_name: the name of the entity date table
        :param label_timespan: the time timespan that labels in matrix will include
        :type label_name: str
        :type label_type: str
        :type entity_date_table_name: str
        :type label_timespan: str
        :return: name of csv containing labels
        :rtype: str
        """
        if self.include_missing_labels_in_train_as is None:
            label_predicate = "r.label"
        elif self.include_missing_labels_in_train_as is False:
            label_predicate = "coalesce(r.label, 0)"
        elif self.include_missing_labels_in_train_as is True:
            label_predicate = "coalesce(r.label, 1)"
        else:
            raise ValueError(
                'incorrect value "{}" for include_missing_labels_in_train_as'.format(
                    self.include_missing_labels_in_train_as
                )
            )

        if self.includes_labels:
            labels_query = self._outer_join_query(
                right_table_name="{schema}.{table}".format(
                    schema=self.db_config["labels_schema_name"],
                    table=self.db_config["labels_table_name"],
                ),
                entity_date_table_name='"{schema}"."{table}"'.format(
                    schema=self.db_config["features_schema_name"],
                    table=entity_date_table_name,
                ),
                right_column_selections=", {} as {}".format(label_predicate, label_name),
                additional_conditions="""AND
                    r.label_name = '{name}' AND
                    r.label_type = '{type}' AND
                    r.label_timespan = '{timespan}'
                """.format(
                    name=label_name, type=label_type, timespan=label_timespan
                ),
                #include_index=False,
                include_index=True,
                column_override=label_name
            )
        else: 
            labels_query = self._join_no_labels_case(
                right_column_selections=f", 0 as {label_name}",
                entity_date_table_name=f'"{self.db_config["features_schema_name"]}"."{entity_date_table_name}"',
            )

        return labels_query
    

    def feature_load_queries(self, feature_dictionary, entity_date_table_name):
        """ Loop over tables in features schema, writing the data from each to a csv. Return the full list of feature 
        csv names and the list of all features.
        :param feature_dictionary: a dictionary of feature tables and features to be included in the matrix
        :param entity_date_table_name: the name of the entity date table for the matrix
        :type feature_dictionary: dict
        :type entity_date_table_name: str
        :return: list of csvs containing feature data
        :rtype: list
        """
        # iterate! for each table, make query
        queries = []
        for num, (feature_table_name, feature_names) in enumerate(feature_dictionary.items()):
            logging.info("Generating feature query for %s", feature_table_name)
            queries.append(self._outer_join_query(
                right_table_name="{schema}.{table}".format(
                    schema=self.db_config["features_schema_name"],
                    table=feature_table_name,
                ),
                entity_date_table_name='{schema}."{table}"'.format(
                    schema=self.db_config["features_schema_name"],
                    table=entity_date_table_name,
                ),
                right_column_selections=[', "{0}"'.format(fn) for fn in feature_names],
            ))
        return queries


    def stitch_csvs(self, features_queries, label_query, matrix_store, matrix_uuid):
        """
        Get all features related this matrix_uuid as CSV files, as well as the labels. 
        Join all the csv elements columnwise and create the final matrix. 
        The last column is the label. 

        Args:
            features_queries (list): List of the requried queries to execute to 
                get all the features from this design matrix. 
            label_query (string): The query required to get the label associated 
                to this design matrix. 
            matrix_store (MatrixStorage): Storage path for the project
            matrix_uuid (string): Id of the matrix

        Returns:
            DataFrame: Design downcasted matrix
        """
        logger.debug(f"stitching csvs for matrix {matrix_uuid}")
        connection = self.db_engine.raw_connection()
        cursor = connection.cursor()
        header = "HEADER"

        # starting with features 
        path_ = str(matrix_store.get_storage_directory())
        logger.debug(f"path to store csvs {path_}")

        filenames = []
        for i, query_string in enumerate(features_queries):
            copy_sql = f"COPY ({query_string}) TO STDOUT WITH CSV {header}"
            bio = io.BytesIO()
            cursor.copy_expert(copy_sql, bio)
            bio.seek(0)
            output_ = bio.read()
            
            filenames.append(path_ + "/" + matrix_uuid + "_" + str(i) + ".csv")
            
            matrix_store.save_tmp_csv(output_, path_, matrix_uuid, f"_{str(i)}.csv")

        logger.debug(f"number of feature files to paste for matrix {matrix_uuid}: {len(filenames)}")

        # label
        copy_sql = f"COPY ({label_query}) TO STDOUT WITH CSV {header}"
        bio = io.BytesIO()
        cursor.copy_expert(copy_sql, bio)
        bio.seek(0)
        output_ = bio.read()

        matrix_store.save_tmp_csv(output_, path_, matrix_uuid, "_label.csv")

        # add label file to filenames
        filenames.append(f"{path_}/{matrix_uuid}_label.csv")
        
        # check if the number of rows among all features and label -if any- files are the same
        try: 
            assert check_rows_in_files(filenames, matrix_uuid)
        except AssertionError as e: 
            logger.exception(
                f"Different number of rows among features and label files for matrix uuid {matrix_uuid} ",
            )
            if self.run_id:
                errored_matrix(self.run_id, self.db_engine)
            raise
    
        # check if the entities_id and knowledge_dates are the same among all the features and label files
        try:
            check_entity_ids_in_files(filenames, matrix_uuid)
        except AssertionError as e:
            logger.exception(
                f"Not the same order of entity id and knowledge date in all features and label files for matrix uuid {matrix_uuid}"
            )
            if self.run_id:
                errored_matrix(self.run_id, self.db_engine)
            raise

        # remove first 2 columns on each features and label files -except the first one- 
        verified_filenames = remove_entity_id_and_knowledge_dates(filenames, matrix_uuid)

        # join all files starting with features and ending with label
        files = " ".join(verified_filenames)

        # save joined csvs
        cmd_line = 'paste ' + files + ' -d "," > ' + path_ + "/" + matrix_uuid + ".csv"
        logger.debug(f"About to paste CSVs columnwise for matrix {matrix_uuid} cmd line: {cmd_line}")
        try:
            subprocess.run(cmd_line, shell=True, check=True)
            logger.debug(f"CSV files {files} pasted into {path_}/{matrix_uuid}.csv")
        except subprocess.CalledProcessError as e:      
            logger.error(f"Error pasting CSV files {files}: {e}")
        
        logger.debug(f"about to load csvmatrix with uuid {matrix_uuid} as polars df")
        start = time.time()
        # load as DF with polars
        filename_ = f'{path_}/{matrix_uuid}.csv'
        #df = pd.read_csv(filename_, parse_dates=["as_of_date"])
        df_pl = pl.read_csv(filename_, infer_schema_length=0).with_columns(pl.all().exclude(
            ['entity_id', 'as_of_date']).cast(pl.Float32, strict=False))
        end = time.time()
        logger.debug(f"time to read csv of matrix with uuid {matrix_uuid} (sec): {(end-start)/60}")
        
        # casting entity_id and as_of_date 
        logger.debug(f"casting entity_id and as_of_date")
        start = time.time()
        # define if as_of_date is date or datetime for correct cast
        if len(df_pl.get_column('as_of_date').head(1)[0].split()) > 1: 
            format = "%Y-%m-%d %H:%M:%S"
        else: 
            format = "%Y-%m-%d"

        df_pl = df_pl.with_columns(pl.col("as_of_date").str.to_datetime(format))
        df_pl = df_pl.with_columns(pl.col("entity_id").cast(pl.Int32, strict=False))
        end = time.time()
        logger.debug(f"time casting entity_id and as_of_date of matrix with uuid {matrix_uuid} (sec): {(end-start)/60}")
        
        logger.debug(f"getting labels pandas series from polars data frame")
        
        # getting label series
        labels_pl = df_pl.select(df_pl.columns[-1])
        # convert into pandas series 
        labels_df = labels_pl.to_pandas()
        labels_series = labels_df.squeeze()

        # remove labels from features and return as df
        logger.debug(f"removing labels from main polars df")
        df_pl_aux = df_pl.drop(df_pl.columns[-1])

        # converting from polars to pandas
        logger.debug(f"about to convert polars df into pandas df")
        start = time.time()
        df = df_pl_aux.to_pandas()
        end = time.time()
        logger.debug(f"Time converting from polars to pandas (sec): {(end-start)/60}")
        df.set_index(["entity_id", "as_of_date"], inplace=True)
        logger.debug(f"df data types: {df.dtypes}")
        logger.debug(f"Pandas DF memory usage: {df.memory_usage(deep=True).sum()/1000000} MB")
        # generating gzip file from csv
        generate_gzip(path_, matrix_uuid)

        # if matrix store is S3 
        store_is_s3 = False
        if isinstance(matrix_store.matrix_base_store, S3Store):
            logger.debug(f"storing {matrix_uuid}.csv.gz on S3")
            matrix_store._save(path_, matrix_store.matrix_base_store.path)
            store_is_s3 = True

        logger.debug(f"removing csvs files for matrix {matrix_uuid}")
        # addinig _sorted and _fixed files to list of files to rm 
        rm_filenames = generate_list_of_files_to_remove(filenames, matrix_uuid)
        remove_unnecessary_files(rm_filenames, path_, matrix_uuid, store_is_s3)

        return df, labels_series

        
        